﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Anwendung                                                                           //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2019 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include "Cy4Net.h"
#include "Cy4TypePlate.h"
#include "Storing.h"
#include "Board.h"
#include "Timer.h"
#include "Application.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                        // Kontaktereignis:
{   CONTACTEVENT_CLOSED,            // · Kontakt wurde geschlossen
    CONTACTEVENT_OPENED,            // · Kontakt wurde geöffnet
    CONTACTEVENT_NONE               // · Kein Ereignis
} ContactEvent;

typedef enum                        // Kontaktzustand:
{   CONTACTSTATE_CLOSED,            // · Kontakt ist geschlossen
    CONTACTSTATE_OPENED,            // · Kontakt ist geöffnet
    CONTACTSTATE_UNKNOWN            // · Kontaktzustand unbekannt (Anfangszustand)
} ContactState;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static  Timer              serverQueryInterval[2] = { TIMER_SECONDS(0), TIMER_SECONDS(0) };
static  Timer              extraOnTimer           = TIMER_SECONDS(0);
static  SInt08             ambientBrightness      = CFG_INVALID_BRIGHTNESS;
static ContactEvent        contactEvent           = CONTACTEVENT_NONE;
static ContactState        contactState           = CONTACTSTATE_UNKNOWN;
static const PushMessage*  currentPushMessage     = NULL;
static const TextFeed*     currentTextFeed        = NULL;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ValidateAmbientBrightness(void)
{
	Cy4Link  serverLink;
	STORING_Load(STO_ServerLink[1], &serverLink);
	bool enabled;
	STORING_Load(STO_AcceptBroadcast[1], &enabled);
	if (!enabled && !CY4LINK_IsValid(&serverLink))
	{   // Keine gültiger Umgebungshelligkeitsserver und Rundrufe werden ignoriert
		ambientBrightness = CFG_INVALID_BRIGHTNESS;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
void
StartExtraOn(void)
{	
	UInt08 duration;
	STORING_Load(STO_ExtraOnDuration, &duration);
	TIMER_Start(&extraOnTimer, duration*60);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
StopExtraOn(void)
{	
	TIMER_Stop(&extraOnTimer);
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
IsExtraOn(void)
{	
	return !TIMER_HasExpired(&extraOnTimer);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Nachrichten-Ausstatter  //////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupContactPushMessage(Cy4Message* message)
{
    PushMessage pushMessage;
    PUSHMESSAGE_EEPLoad(currentPushMessage, &pushMessage);
    PUSHMESSAGE_ToMessage(&pushMessage, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupContactTextFeed(Cy4Message* message)
{
    TextFeed textFeed;
    TEXTFEED_EEPLoad(currentTextFeed, &textFeed);
    TEXTFEED_ToMessage(&textFeed, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupQueryTimeMessage(Cy4Message* message)
{
	Cy4Link serverLink;
	STORING_Load(STO_ServerLink[0], &serverLink);
	if (CY4LINK_IsValid(&serverLink))
	{   CY4MESSAGE_SetupRead(message, serverLink.address, serverLink.location, 7);
	}
	else
	{   CY4MESSAGE_Invalidate(message);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupQueryBrightnessMessage(Cy4Message* message)
{
	Cy4Link serverLink;
	STORING_Load(STO_ServerLink[1], &serverLink);
	if (CY4LINK_IsValid(&serverLink))
	{   CY4MESSAGE_SetupRead(message, serverLink.address, serverLink.location, 1);
	}
	else
	{   CY4MESSAGE_Invalidate(message);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Antwort-Behandler  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ResponseOfQueryTime(const Cy4Message* message)
{
	if (CY4MESSAGE_IsValid(message))
	{   DateTime dateTime;
		DATETIME_Deserialize(CY4MESSAGE_ConstData(message), &dateTime);
		if (DATETIME_IsValid(&dateTime)) BOARD_SetClock(&dateTime);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ResponseOfQueryBrightness(const Cy4Message* message)
{
	if (CY4MESSAGE_IsValid(message))
	{   ambientBrightness = (SInt08)CY4MESSAGE_ConstData(message)[0];
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Anfrage-Bearbeiter  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: SwitchMode (read write)                                                     //
//                                                                                      //
//  Beschreibung: Betriebsart des Schaltausgangs.                                       //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |PSM|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//    [PSM] Betriebsart: 0 = Aus (Voreinstellung),                                      //
//                       1 = Ein,                                                       //
//                       2 = Zeitautomatik,                                             //
//                       3 = Helligkeitsautomatik,                                      //
//                       x = Wird ignoriert                                             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnSwitchMode(const Cy4Service* service, Cy4Message* message)
{
	SwitchMode mode;
	STORING_Load(STO_PowerSwitchMode, &mode);
	UInt08 data = (UInt08)mode;
	CY4SERVICE_TransferPayload(service, message, &data);
	if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
	{   if (data < 4)
		{   mode = (SwitchMode)data;
			STORING_Save(&mode, STO_PowerSwitchMode);			
			StopExtraOn();
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: SwitchState (read only)                                                     //
//                                                                                      //
//  Beschreibung: Liefert den Zustand des Schaltausgangs.                               //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |PSS|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//    [PSS] Schaltausgang: 0 = Aus, 1 = Ein.                                            //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnSwitchState(const Cy4Service* service, Cy4Message* message)
{
	(void)service;	
	CY4MESSAGE_Data(message)[0] = BOARD_Switch();
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ExtraOn (write only)                                                        //
//                                                                                      //
//  Beschreibung: Auslöser, um unabhängig von der aktuellen Betriebsart den Schalt-     //
//                ausgang für eine bestimmte Dauer (s. ExtraOnDuration.DUR) einzu-      //
//                schalten.																//
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EXO|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//                                                                                      //
//  [EXO] Auslöser zur betriebsartunabhängigen zeitbegrenzten Einschaltung des          //
//        Schaltausgangs: 5Eh ('^').                                                    //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnExtraOn(const Cy4Service* service, Cy4Message* message)
{
	(void)service;	
	if (CY4MESSAGE_ConstData(message)[0] == 0x5E) StartExtraOn();
	else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ExtraOnDuration (read write only)                                           //
//                                                                                      //
//  Beschreibung: Dauer der Extraeinschaltung. Siehe ExtraOn.EXO.                       //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |DUR|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//                                                                                      //
//  [DUR] Dauer der betriebsartunabhängigen Einschaltung des Schaltausgangs: 1 ..		//
//        240 [min]																		//
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnExtraOnDuration(const Cy4Service* service, Cy4Message* message)
{	
	UInt08 data;
	STORING_Load(STO_ExtraOnDuration, &data);
	CY4SERVICE_TransferPayload(service, message, &data);
	if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
	{   if (data <   1) data =   1;
		if (data > 240) data = 240;
		STORING_Save(&data, STO_ExtraOnDuration);		
		if (IsExtraOn()) StartExtraOn();
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: TimerModeParameter (read write)                                             //
//                                                                                      //
//  Beschreibung: Zugriff auf die Parameter der Zeitautomatik. Überschneiden sich       //
//                Zeiten der Zeitgeber, dominiert der eingeschaltete Zustand.           //
//                TimerModeParameter ist ein Gemeinschaftsbearbeiter für alle 3 Zeit-   //
//                geber.                                                                //
//                                                                                      //
//              0   1   2   3   4                                                       //
//            +---+---+---+---+---+  5   6   7   8   9   10  11                         //
//  Belegung: |ACT|STH|STM|ETH|ETM|+---+---+---+---+---+---+---+                        //
//            +---+---+---+---+---+|MON|TUE|WED|THU|FRI|SAT|SUN|                        //
//                                 +---+---+---+---+---+---+---+                        //
//                                                                                      //
//    [ACT] Zeitgeber Aktiv: 0 = Aus, 1 = Ein, x = Wird ignoriert.                      //
//    [STH] Einschaltzeitpunkt Stunde: 0 .. 23 [h], oder 255.                           //
//    [STM] Einschaltzeitpunkt Minute: 0 .. 59 [m], oder 255.                           //
//    [ETH] Abschaltzeitpunkt Stunde:  0 .. 23 [h], oder 255.                           //
//    [ETM] Abschaltzeitpunkt Minute:  0 .. 59 [m], oder 255.                           //
//    [MON] Freigabe für Montag:     0 = Aus 1 = Ein, x = Wird ignoriert.               //
//    [TUE] Freigabe für Dienstag:   0 = Aus 1 = Ein, x = Wird ignoriert.               //
//    [WED] Freigabe für Mittwoch:   0 = Aus 1 = Ein, x = Wird ignoriert.               //
//    [THU] Freigabe für Donnerstag: 0 = Aus 1 = Ein, x = Wird ignoriert.               //
//    [FRI] Freigabe für Freitag:    0 = Aus 1 = Ein, x = Wird ignoriert.               //
//    [SAT] Freigabe für Samstag:    0 = Aus 1 = Ein, x = Wird ignoriert.               //
//    [SUN] Freigabe für Sonntag:    0 = Aus 1 = Ein, x = Wird ignoriert.               //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnTimerModeParameter(const Cy4Service* service, Cy4Message* message)
{	
	UInt08 n = ((CY4MESSAGE_Location(message) & 0x0000F0)>>4)-2; // n. Zeitgeber
	AutoTimer autoTimer;
	STORING_Load(STO_AutoTimer[n], &autoTimer);
	UInt08 data[12];
	data[0] = (UInt08)autoTimer.enabled;
	data[1] = autoTimer.startTime.hour;
	data[2] = autoTimer.startTime.minute;
	data[3] = autoTimer.endTime.hour;
	data[4] = autoTimer.endTime.minute;
	for (UInt08 d=0; d<7; d++) data[5+d] = (UInt08)autoTimer.weekDay[d];
	CY4SERVICE_TransferPayload(service, message, data);
	if (data[0] <  2) autoTimer.enabled = (bool)data[0];
	if (data[1] < 24) autoTimer.startTime.hour   = data[1];
	if (data[2] < 60) autoTimer.startTime.minute = data[2];
	autoTimer.startTime.second = 0;
	if (data[3] < 24) autoTimer.endTime.hour   = data[3];
	if (data[4] < 60) autoTimer.endTime.minute = data[4];
	autoTimer.endTime.second = 0;
	for (UInt08 d=0; d<7; d++)
	{   if (data[5+d] < 2) autoTimer.weekDay[d] = (bool)data[5+d];
	}
	STORING_Save(&autoTimer, STO_AutoTimer[n]);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: BrightnessModeParameter (read write)                                        //
//                                                                                      //
//  Beschreibung: Zugriff auf die Parameter der Helligkeitsautomatik.                   //
//                                                                                      //
//              0   1   2   3                                                           //
//            +---+---+---+---+                                                         //
//  Belegung: |ACT|THB|THD|ONT|                                                         //
//            +---+---+---+---+                                                         //
//                                                                                      //
//    [ACT] Aktivität                                                                   //
//          0 = Ein wenn dunkel, aus wenn hell (Voreinstellung).                        //
//          1 = Ein wenn hell, aus wenn dunkel.                                         //
//          x = Ohne Wirkung                                                            //
//    [THB] Helligkeits-Schwellenwert: 2 .. 99 %. Wert muss über <THD> liegen.          //
//    [THD] Dunkelheits-Schwellenwert: 1 .. 98 %. Wert muss unter <THB> liegen.         //
//    [ONT] Automatische Abschaltung nach (in 10-Minuten Schritten): 0 .. 96 [10m],     //
//          0 = Keine Abschaltung .                                                     //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnBrightnessModeParameter(const Cy4Service* service, Cy4Message* message)
{
	UInt08 data[5];	
	BrightnessMode mode;
	STORING_Load(STO_BrightnessMode, &mode);
	data[0] = (UInt08)mode;
	STORING_Load(STO_LightnessThreshold, &data[1]);
	STORING_Load(STO_DarknessThreshold, &data[2]);
	STORING_Load(STO_BrightnessAutoOffTime, &data[3]);
	CY4SERVICE_TransferPayload(service, message, data);
	if (data[0] < 2)
	{   mode = (BrightnessMode)data[0];
		STORING_Save(&mode, STO_BrightnessMode);
	}
	if (data[2] > data[1]) data[2] = data[1]-1;
	if (data[1] <  2) data[1] =  2;
	if (data[1] > 99) data[1] = 99;
	STORING_Save(&data[1], STO_LightnessThreshold);
	if (data[2] <  1) data[2] =  1;
	if (data[2] > 98) data[2] = 98;
	STORING_Save(&data[2], STO_DarknessThreshold);
	if (data[4] > 96) data[4] = 96;
	STORING_Save(&data[3], STO_BrightnessAutoOffTime);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: DateTime (read write)                                                       //
//                                                                                      //
//  Beschreibung: Ermöglicht den Zugriff auf die Systemzeit des Gerätes. Der Behandler  //
//                wird für den lokalen Dienst als auch für den Rundruf eingesetzt.      //
//                                                                                      //
//              0   1   2   3   4   5   6                                               //
//            +---+---+---+---+---+---+---+                                             //
//  Belegung: |YR1|YR2|MON|DAY|HUR|MIN|SEC|                                             //
//            +---+---+---+---+---+---+---+                                             //
//                                                                                      //
//    [YRx] Jahr:    1700 .. 9000 , (0xFFFF)                                            //
//    [MON] Monat:      1 ..   12 , (0xFF)                                              //
//    [DAY] Tag:        1 ..   31 , (0xFF)                                              //
//    [HUR] Stunde:     0 ..   23 , (0xFF)                                              //
//    [MIN] Minute:     0 ..   59 , (0xFF)                                              //
//    [SEC] Sekunde:    0 ..   59 , (0xFF)                                              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnDateTime(const Cy4Service* service, Cy4Message* message)
{
	if (CY4MESSAGE_Location(message) == 0xE00020)
	{   bool broadcastAccepted;
		STORING_Load(STO_AcceptBroadcast[0], &broadcastAccepted);
		if (!broadcastAccepted) return;
	}
	UInt08 data[7];
	{   DateTime  dateTime;
		BOARD_Clock(&dateTime);
		DATETIME_Serialize(&dateTime, data);
	}
	CY4SERVICE_TransferPayload(service, message, data);
	if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
	{   DateTime dateTime;
		DATETIME_Deserialize(data, &dateTime);
		BOARD_SetClock(&dateTime);
		TIMER_Start(&serverQueryInterval[0], CFG_TIME_QUERY_INTERVAL*60);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: AmbientBrightness (read write)                                              //
//                                                                                      //
//  Beschreibung: Ermöglicht Zugriff auf die vom System genutzte Umgebungshelligkeit.   //
//                Der Behandler wird für den lokalen Dienst als auch für den Rundruf    //
//                eingesetzt.                                                           //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EBS|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EBS] Umgebungshelligkeit: 0 .. 100 % oder -128.                                    //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnAmbientBrightness(const Cy4Service* service, Cy4Message* message)
{
	if (CY4MESSAGE_Location(message) == 0xE00060)
	{   bool broadcastAccepted;
		STORING_Load(STO_AcceptBroadcast[1], &broadcastAccepted);
		if (!broadcastAccepted) return;
	}
	UInt08 data = ambientBrightness;
	CY4SERVICE_TransferPayload(service, message, &data);
	ambientBrightness = data;
	if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
	{   TIMER_Start(&serverQueryInterval[1], CFG_BRIGHTNESS_QUERY_INTERVAL*60);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: AcceptBroardcasts (read write)                                              //
//                                                                                      //
//  Beschreibung: Mit diesem Dienst kann festgelegt werden, ob das Gerät Rundmitteil-   //
//                ungen an E00020h (Zeit) und E00060h (Umgebungshelligkeit) entgegen-   //
//                nimmt oder nicht.                                                     //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |ATB|ABB|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [ATB] Zeit-Rundmitteilung: 1 = akzeptieren, 0 = ignorieren.                         //
//  [ABB] Umgebungshelligkeit-Rundmitteilung: 1 = akzeptieren, 0 = ignorieren.          //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnAcceptBroardcasts(const Cy4Service* service, Cy4Message* message)
{
	UInt08 data[2];
	for (UInt08 i=0; i<2; i++)
	{   bool enable;
		STORING_Load(STO_AcceptBroadcast[i], &enable);
		data[i] = (UInt08)enable;
	}
	CY4SERVICE_TransferPayload(service, message, data);
	for (UInt08 i=0; i<2; i++)
	{   bool enable = (bool)data[i];
		STORING_Save(&enable, STO_AcceptBroadcast[i]);
	}
	if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
	{   ValidateAmbientBrightness();
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ServerLink (read write)                                                     //
//                                                                                      //
//  Beschreibung: Adresse und Dienstort der Server für Zeit und Umgebungshelligkeit.    //
//                                                                                      //
//              0   1   2   3   4   5   6   7   8   9                                   //
//            +---+---+---+---+---+---+---+---+---+---+                                 //
//  Belegung: |TSN|TSD|LC1|LC2|LC3|BSN|BSD|EC1|EC2|EC3|                                 //
//            +---+---+---+---+---+---+---+---+---+---+                                 //
//                                                                                      //
//  [TSN] Zeitserver: Cy4-Netzadresse: 1 .. 255.                                        //
//  [TSD] Zeitserver: Cy4-Geräteadresse: 1 .. 255.                                      //
//  [LCx] Zeitserver: Dienstort der abzufragenden Zeitinformation: 100000h .. FFFF00h,  //
//        FFFFFFh wenn nicht gesetzt.                                                   //
//  [BSN] Umgebungshelligkeitsserver: Cy4-Netzadresse: 1 .. 255.                        //
//  [BSD] Umgebungshelligkeitsserver: Cy4-Geräteadresse: 1 .. 255.                      //
//  [ECx] Umgebungshelligkeitsserver: Dienstort der abzufragenden helligkeit: 100000h   //
//        .. FFFF00h, FFFFFFh wenn nicht gesetzt .                                      //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnServerLink(const Cy4Service* service, Cy4Message* message)
{
	UInt08  data[10];
	for (UInt08 i=0; i<2; i++)
	{   Cy4Link  serverLink;
		STORING_Load(STO_ServerLink[i], &serverLink);
		CY4LINK_ToData(&serverLink, &data[5*i]);
	}
	CY4SERVICE_TransferPayload(service, message, data);
	for (UInt08 i=0; i<2; i++)
	{   Cy4Link  serverLink;
		CY4LINK_FromData(&data[5*i], &serverLink);
		STORING_Save(&serverLink, STO_ServerLink[i]);
		if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
		{   if (CY4LINK_IsValid(&serverLink)) TIMER_Start(&serverQueryInterval[i], 3);
			ValidateAmbientBrightness();
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ContactState (read only)                                                    //
//                                                                                      //
//  Beschreibung: Ermittelt den Zustand des Kontaktes unter Brücksichtigung der Er-     //
//  eignis-Verzögerungen (s. ContactEventConfiguration).                                //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |C1S|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [CnS] Zustand Kontakt n: 0|1 = geöffnet|geschlossen.                                //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnContactState(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data;
   
    bool closed;
    switch (contactState)
    {   case CONTACTSTATE_OPENED:  closed = false; break;
        case CONTACTSTATE_CLOSED:  closed = true;  break;
        default:                   closed = false;    
        data = closed;
	}
    CY4SERVICE_TransferPayload(service, message, &data);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ContactStateSignal (read write)                                             //
//                                                                                      //
//  Beschreibung: Legt die Art des Licht- und Tonsignals fest, das bei Zustand des      //
//                entsprechenden Kontaktes ausgegeben wird. Der Kontaktzustand wird     //
//                unter Berücksichtigung der Eregniskonfiguration ermittelt (siehe      //
//                Dienst ContactEventConfiguration).                                    //
//                                                                                      //
//              0   1   2   3   4   5                                                   //
//            +---+---+---+---+---+---+                                                 //
//  Belegung: |TSC|LSC|BRC|TSO|LSO|BRO|                                                 //
//            +---+---+---+---+---+---+                                                 //
//                                                                                      //
//                                                                              	    //
//                                                                              		//
//                                                                              		//
//                                                                              		//
//                                                                              		//
//                                                                                      //
//  [TSC] Tonsignal, wenn Kontakt geschlossen ist.                                      //
//          0 = Aus (Voreinstellung)                                                    //
//          1 = Tropfen                                                                 //
//          2 = Ticken                                                                  //
//          3 = Martinticks                                                             //
//          4 = Kurzer Piepton D7                                                       //
//          5 = Langer Piepton D7                                                       //
//          * = Aus                                                                     //
//  [LSC] Lichtsignal, wenn Kontakt geschlossen ist.                                    //
//          0 = Kein Signal, LED aus (Voreinstellung).                                  //
//          1 = Ein                                                                     //
//          2 = Blinkt                                                                  //
//          3 = Blitzt                                                                  //
//          4 = Pulsiert                                                                //
//          5 = Herzschlag                                                              //
//          * = Kein Signal, LED aus.                                                   //
//  [BRC] Helligkeit der LED bei Lichtsignal [LSC]: 10 .. 100 %					        //
//  [TSO] Tonsignal, wenn Kontakt geöffnet ist (Wert wie bei TSC)                       //
//  [LSO] Lichtsignal, wenn Kontakt geöffnet ist (Wert wie bei LSC).                    //
//  [BRO] Helligkeit der LED bei Lichtsignal [LSO]: 10 .. 100 %					        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnContactStateSignal(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[6];    
    {	ContactStateSignal signal;
		// Geschlossen
		STORING_Load(STO_ContactStateSignal[0], &signal);
		data[0] = (UInt08)signal.toneSignal;
		data[1] = (UInt08)signal.lightSignal;
		data[2] = (UInt08)signal.brightness;
		// Geöffnet
		STORING_Load(STO_ContactStateSignal[1], &signal);
		data[3] = (UInt08)signal.toneSignal;
		data[4] = (UInt08)signal.lightSignal;
		data[5] = (UInt08)signal.brightness;    
	}
    CY4SERVICE_TransferPayload(service, message, data);    
    {   ContactStateSignal signal;
        // Geschlossen
        if (data[0] < (UInt08)TONESTATESIGNAL_NUMBER)  signal.toneSignal  = (ToneStateSignal)data[0];
        else                                           signal.toneSignal  = TONESTATESIGNAL_OFF;
        if (data[1] < (UInt08)LIGHTSTATESIGNAL_NUMBER) signal.lightSignal = (LightStateSignal)data[1];
        else                                           signal.lightSignal = LIGHTSTATESIGNAL_OFF;		
		if      (data[2] <  10) signal.brightness =  10;
		else if (data[2] > 100) signal.brightness = 100;
		else                    signal.brightness = (UInt08)data[2];		
        STORING_Save(&signal, STO_ContactStateSignal[0]);
        // Geöffnet
        if (data[3] < (UInt08)TONESTATESIGNAL_NUMBER)  signal.toneSignal  = (ToneStateSignal)data[3];
        else                                           signal.toneSignal  = TONESTATESIGNAL_OFF;
        if (data[4] < (UInt08)LIGHTSTATESIGNAL_NUMBER) signal.lightSignal = (LightStateSignal)data[4];
        else                                           signal.lightSignal = LIGHTSTATESIGNAL_OFF;
		if      (data[5] <  10) signal.brightness =  10;
		else if (data[5] > 100) signal.brightness = 100;
		else                    signal.brightness = (UInt08)data[5];
        STORING_Save(&signal, STO_ContactStateSignal[1]);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ContactEventConfiguration (read write)                                      //
//                                                                                      //
//  Beschreibung: Bestimmt Verhalten und Charakteristik des jeweiligen Kontakte-        //
//                wechsels. ContactEventConfiguration ist ein Gemeinschaftsbearbeiter   //
//                für folgende Dienste:                                                 //
//                                                                                      //
//                  130100h:61: Parameter bei Ereignis Kontakt geschlossen              //
//                  130180h:61: Parameter bei Ereignis Kontakt geöffnet                 //
//                                                                                      //
//              0         2   3                             10  11  12        59  60    //
//            +---+  1  +---+---+  4   5   6   7   8   9  +---+---+---+-   -+---+---+   //
//  Belegung: |DUR|+---+|LSE|BRT|+---+---+---+---+---+---+|T00|T01|T02| ... |T49| 0 |   //
//            +---+|TSE|+---+---+|NET|DEV|LC0|LC1|LC2|PAR|+---+---+---+-   -+---+---+   //
//                 +---+         +---+---+---+---+---+---+                              //
//                                                                                      //
//  [DUR] Zustandsdauer. Bestimmt die Dauer, wie lange der jeweilige Kontaktzustand     //
//        anhalten muss, bis das entsprechende Ereignis eintritt: 0 .. 240 [s].         //
//  [TSE] Akustisches Signal des Piezosignalgebers bei o.g. Ereignis:                   //
//           0 = Kein Signal (Voreinstellung).                                          //
//           1 = Tief Hoch                                                              //
//           2 = Hoch Tief                                                              //
//           3 = Tief Mitte Hoch                                                        //
//           4 = Hoch Mitte Tief                                                        //
//           5 = Taste gedrückt                                                         //
//           6 = Oktave2Hoch                                                            //
//           7 = Doppelton                                                              //
//           8 = Trippelton                                                             //
//           9 = Zwitscher                                                              //
//          10 = Kurzer Piepton C6                                                      //
//          11 = Langer Piepton C6                                                      //
//          12 = Kurzer Piepton C8                                                      //
//          13 = Langer Piepton C8                                                      //
//          14 = Kurzer Piepton C9                                                      //
//          15 = Langer Piepton C9                                                      //
//           * = Kein Signal, Stille                                                    //
//  [LSE] Lichtsignaleffekt der LED bei o.g. Ereignis:                                  //
//           0 = LED aus (Voreinstellung).                                              //
//           1 = Blinkt einmal                                                          //
//           2 = Blinkt zweimal                                                         //
//           3 = Blinkt dreimal                                                         //
//           4 = Blinkt sechsmal                                                        //
//           5 = Blitzt einmal                                                          //
//           6 = Blitzt zweimal                                                         //
//           7 = Blitzt dreimal                                                         //
//           8 = Blitzt sechsmal                                                        //
//           9 = Einblenden                                                             //
//          10 = Ausblenden                                                             //
//          11 = Flackern                                                               //
//           * = Kein Signal, LED aus.                                                  //
//  [BRT] Helligkeit der LED beim Lichtsignaleffekt [LSE]: 10 .. 100 %					//
//  [NET] Push-Nachricht Ziel bei o.g. Ereignis, Netzadresse   : 0 ..     255           //
//  [DEV] Push-Nachricht Ziel bei o.g. Ereignis, Geräteadresse : 0 ..     255           //
//  [LCx] Push-Nachricht Ziel bei o.g. Ereignis, Dienstort     : 0 .. 0xFFFFF           //
//  [PAR] Push-Nachricht Ziel bei o.g. Ereignis, Parameter-Byte: 0 ..     255           //
//  [Txx] Zeichen der Rundmitteilung bei o.g. Ereignis: 0 .. 255                        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnContactEventConfiguration(const Cy4Service* service, Cy4Message* message)
{
    UInt32 location = CY4MESSAGE_Location(message);
    void* eepAddress;
    if     ( location < 0x130180 ) eepAddress = STORING_AddressOf(STO_ContactEventConfiguration[0]);
    else                           eepAddress = STORING_AddressOf(STO_ContactEventConfiguration[1]);    
    ContactEventConfiguration configuration;
    STORING_LoadEEPMem(eepAddress, &configuration, sizeof(ContactEventConfiguration));
    UInt08  data[61];
    data[0] = configuration.duration;
    data[1] = (UInt08)configuration.toneSignal;
    data[2] = (UInt08)configuration.lightSignal;
	data[3] = (UInt08)configuration.brightness;
    PUSHMESSAGE_ToData(&configuration.pushMessage, &data[4]);
    TEXTFEED_ToData(&configuration.textFeed, &data[10]);
    CY4SERVICE_TransferPayload(service, message, data);
    if (data[0] > 240) configuration.duration = 240;
    else               configuration.duration = data[0];
    if (data[1] < (UInt08)TONEEVENTSIGNAL_NUMBER) configuration.toneSignal = (ToneEventSignal)data[1];
    else                                          configuration.toneSignal = TONEEVENTSIGNAL_NONE;
    if (data[2] < (UInt08)LIGHTEVENTSIGNAL_NUMBER) configuration.lightSignal = (LightEventSignal)data[2];
    else                                           configuration.lightSignal = LIGHTEVENTSIGNAL_NONE;
	if      (data[3] <  10) configuration.brightness =  10;
	else if (data[3] > 100) configuration.brightness = 100;
	else                    configuration.brightness = (UInt08)data[3];	
    PUSHMESSAGE_FromData(&data[4], &configuration.pushMessage);
    TEXTFEED_FromData(&data[10], &configuration.textFeed);
    STORING_SaveEEPMem(&configuration, eepAddress, sizeof(ContactEventConfiguration));
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ContactEventTest (write only)                                               //
//                                                                                      //
//  Beschreibung: Löst zu Probewecken die eingestellte Kontakt-Ereignis-Signalisierung	//
//				  aus.																    //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |CEv|OEv|                                                                 //
//            +---+---+                                                                 //
//					  												                    //
//                                                                                      //
//  [CEv] Auslöser der Signalisierung Ereignis Kontakt geschlossen: 94 (5Eh, '^').      //
//  [OEv] Auslöser der Signalisierung Ereignis Kontakt geöffnet: 94 (5Eh, '^').         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnContactEventTest(const Cy4Service* service, Cy4Message* message)
{
	UInt08  data[] = { 0, 0 };
	CY4SERVICE_TransferPayload(service, message, data);
	UInt08 i = CARDINALITY(data);
	do
	{	i--;
		if (data[i] == 0x5E)
		{	UInt08 ce = i&1;	// Kontakt-Ereignis: 0 = geschlossen, 1 = geöffnet
			ContactEventConfiguration configuration;			
			STORING_Load(STO_ContactEventConfiguration[ce], &configuration);
			LIGHTSIGNAL_Play(configuration.lightSignal, configuration.brightness);
			TONESIGNAL_Play(configuration.toneSignal);
			break; // Auslösen mehrerer Tests ergibt keinen Sinn: Bearbeiter verlassen
		}
	} while (i != 0);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: SystemLedSignal (read write)                                                //
//                                                                                      //
//  Beschreibung: Legt die Aufgabe der System-System-LEDs fest. Folgende Lichtsignale   //
//                können eingestellt werden:                                            //
//                                                                                      //
//                       0 = Kein Signal, Ausgeschaltet.                                //
//                       1 = Leuchtet durchgehend.                                      //
//                       2 = Herzschlag-Signal.                                         //
//                       3 = Zeigt Aktivität auf dem Bus.                               //
//                       4 = Zeigt Herzschlag-Signal und Bus-Aktivität (2 & 3).			//
//                       x = Wirkungslos.                                               //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |SIG|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [SIG] Lichtsignale der System-LED, Voreingestellt ist 3.                            //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnSystemLedSignal(const Cy4Service* service, Cy4Message* message)
{
	UInt08 data[1];
	{   SystemLedSignal signal;
		STORING_Load(STO_SystemLedSignal, &signal);
		data[0] = (UInt08)signal;
	}
	CY4SERVICE_TransferPayload(service, message, data);
	if (data[0] < 5)
	{   SystemLedSignal signal = (SystemLedSignal)data[0];
		STORING_Save(&signal, STO_SystemLedSignal);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: BusActivityOnExternalLed (read write)                                       //
//                                                                                      //
//  Beschreibung: Bestimmt, ob und mit welcher Helligkeit Bus-Aktivität mit dem Gerät   //
//                auf der externen LED angezeigt wird.                                  //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |ACT|BRT|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [ACT] Bus-Aktivität wird auf der LED angezeigt: 0 = Nein, sonst Ja.					//
//  [BRT] Helligkeit der LED bei Bus-Aktivität: 10 .. 100 %			                    //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnBusActivityOnExternalLed(const Cy4Service* service, Cy4Message* message)
{    	
	LedBusActivity ledBusActivity;
    STORING_Load(STO_LedBusActivity, &ledBusActivity);
	UInt08 data[2];
	data[0] = (bool)ledBusActivity.enabled;
	data[1] = ledBusActivity.brightness;
    CY4SERVICE_TransferPayload(service, message, data);
	ledBusActivity.enabled = (bool)data[0];	
	ledBusActivity.brightness = data[1];
	if      (data[1] <  10) ledBusActivity.brightness =  10;
	else if (data[1] > 100) ledBusActivity.brightness = 100;
	else                    ledBusActivity.brightness = data[1];  
	STORING_Save(&ledBusActivity, STO_LedBusActivity);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: DeviceLabel (read write)                                                    //
//                                                                                      //
//  Beschreibung: Benutzerdefinierte Gerätebezeichnung.                                 //
//                                                                                      //
//              0   1   2   ...  59  60                                                 //
//            +---+---+---+-   -+---+---+                                               //
//  Belegung: |N01|N02|N03| ... |N60| 0 |                                               //
//            +---+---+---+-   -+---+---+                                               //
//                                                                                      //
//  [Nxx] Zeichen der allgemeinen Gerätebezeichnung: 0 .. 255 (ASCII/UTF-8)             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnDeviceLabel(const Cy4Service* service, Cy4Message* message)
{
    UInt08  label[61];

    STORING_Load(STO_DeviceLabel, label);
    CY4SERVICE_TransferPayload(service, message, label);
    label[sizeof(label)-1] = '\0';
    STORING_Save(label, STO_DeviceLabel);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: FactorySettings (write only)                                                //
//                                                                                      //
//  Beschreibung: Zurückstellen aller Einstellungen auf den Ursprungszustand.           //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |FST|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [FST] Auslöser zur Wiederherstellung der Werksvoreinstellungen: 94 (5Eh, '^').      //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnFactorySettings(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) STORING_RestoreDefaults();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: CallTypePlate (read only)                                            //
//                                                                                      //
//  Beschreibung: Geräte-Typenschild erfragen.                                          //
//                                                                                      //
//              0   1   2   3   4   5   6                                               //
//            +---+---+---+---+---+---+---+                                             //
//  Belegung: |AT2|AT1|AT0|AVE|BT2|BT1|BT0|                                             //
//            +---+---+---+---+---+---+---+                                             //
//                                                                                      //
//  [ATx] Anwendungstypennummer: Axxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//  [AVE] Anwendungsversion:     01 .. 99, Haupt/Unternummer in BCD (Bsp: 12h = V1.2)   //
//  [BTx] Boardtypennummer:      Bxxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//                                                                                      //
//                                                                                      //
//      Beispiel:                                                                       //
//                                                                                      //
//          +---+---+---+---+---+---+---+                                               //
//          |A0h|01h|23h|45h|B0h|67h|89h|      Anwendungstypennummer: 123               //
//          +---+---+---+---+---+---+---+      Anwendungsversion:     4.5               //
//                                             Boardtypennummer:      6789              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnCallTypePlate(const Cy4Service* service, Cy4Message* message)
{
    static const UInt08  typePlate[][7] PROGMEM = {  { 'C','y','4','C','o','d','e' },
                                                     CY4TYPEPLATE(DEV_APPL_TYPE,
                                                                  DEV_APPL_VERS,
                                                                  DEV_BOARD_TYPE)
                                                  };

    UInt08 j = (UInt08)(CY4MESSAGE_Location(message) - CY4SERVICE_Base(service));
    for (UInt08 i=0; i<CY4MESSAGE_Count(message); i++)
    {   CY4MESSAGE_Data(message)[i] = pgm_read_byte(&typePlate[1][j]);
        j++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterApplication (write only)                                        //
//                                                                                      //
//  Beschreibung: Startet die Anwendung neu.                                            //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EAP|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EAP] Auslöser zum (Neu-)Start der Applikation (Reset): 94 (5Eh, '^').              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterApplication(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_Reset();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterProgramLoader (write only)                                      //
//                                                                                      //
//  Beschreibung: Startet das System neu und kehrt im Programmlader-Modul zurück.       //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EPL|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EPL] Auslöser zum Eintritt in den Programmlader: 94 (5Eh, '^').                    //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterProgramLoader(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_EnterPgmLdr();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: ChangeCy4Address (write only)                                        //
//                                                                                      //
//  Beschreibung: Wechsel der Geräteadresse.                                            //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |NET|DEV|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [NET] Neue Netzadresse:   1 .. 255                                                  //
//  [DEV] Neue Geräteadresse: 1 .. 255                                                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnChangeCy4Address(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[2];
    {   Cy4Address address = CY4NET_Address();
        data[0] = CY4ADDRESS_Net(&address);
        data[1] = CY4ADDRESS_Dev(&address);
    }
    CY4SERVICE_TransferPayload(service, message, data);
    {   Cy4Address address = CY4ADDRESS(data[0], data[1]);
        if (CY4ADDRESS_IsBroadcast(&address) || !CY4ADDRESS_IsValid(&address))
        {   CY4MESSAGE_Invalidate(message);
        }
        else
        {   CY4NET_SetAddress(address);
            STORING_SaveCy4SelfAdr(address);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProccedServerRequests(void)
{
	if (TIMER_HasExpired(&serverQueryInterval[0]))
	{   TIMER_Start(&serverQueryInterval[0], CFG_TIME_QUERY_INTERVAL*60);
		CY4NET_Post(SetupQueryTimeMessage, ResponseOfQueryTime);
	}
	if (TIMER_HasExpired(&serverQueryInterval[1]))
	{   TIMER_Start(&serverQueryInterval[1], CFG_BRIGHTNESS_QUERY_INTERVAL*60);
		CY4NET_Post(SetupQueryBrightnessMessage, ResponseOfQueryBrightness);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ProceedContacting(): Verfahrensfunktion der Kontaktierung. Die Zustand des Auto-    //
//  -------------------  maten Automaten findet sich in <contactState>.                 //
//                                                                                      //
//   Zustandsübergangsdiagramm:                                                         //
//                                                   PinState=false,                    //
//                                      +----------+ StartTimer()                       //
//                  +------------------>|  OPENED  +-----------------+                  //
//                  | State: Opened     +----------+                 |                  //
//                  | Event: Opened           /|\                    |                  //
//                  | TIMEOUT                  |                    \|/                 //
//            +-----+-----+                    |  PinState=true +---------+             //
//            |  OPENING  +-----------------+  +----------------+ CLOSING |             //
//            +-----------+ PinState=false  |                   +----+----+             //
//                 /|\                      |                        | TIMEOUT          //
//                  |                      \|/                       | State: Closed    //
//                  |                   +----------+                 | Event: Closed    //
//                  +-------------------|  CLOSED  |<----------------+                  //
//                      PinState=true,  +----------+                                    //
//                      StartTimer()                                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedContacting(void)
{
	typedef enum     // Kontaktierungs-Zustände:
	{   OPENED,      // · Kontakt ist geöffnet
		CLOSING,     // · Kontakt schliessend
		CLOSED,      // · Kontakt ist geschlossen
		OPENING,     // · Kontakt öffnend
		UNKNOWN      // · Kontaktzustand unbekannt (Anfangszustand)
	} State;

	static Timer durationTimer = TIMER_SECONDS(0);
	static State state         = UNKNOWN;
	
	contactEvent = CONTACTEVENT_NONE;     // Ereignis existiert nur einen Durchlauf
	bool isOpen  = BOARD_PinState();
	switch (state)
	{   case OPENED:
		if( !isOpen )
		{   UInt08 duration;
			STORING_Load(STO_ContactEventConfiguration[0].duration, &duration);
			TIMER_Start(&durationTimer, duration);
			state = CLOSING;
		}
		break;
		case CLOSING:
		if( isOpen )
		{   state = OPENED;
		}
		else if( TIMER_HasExpired(&durationTimer) )
		{   state = CLOSED;
			contactState = CONTACTSTATE_CLOSED;
			contactEvent = CONTACTEVENT_CLOSED;
		}
		break;
		case CLOSED:
		if( isOpen )
		{   UInt08 duration;
			STORING_Load(STO_ContactEventConfiguration[1].duration, &duration);
			TIMER_Start(&durationTimer, duration);
			state = OPENING;
		}
		break;
		case OPENING:
		if( !isOpen )
		{   state = CLOSED;
		}
		else if( TIMER_HasExpired(&durationTimer) )
		{   state = OPENED;
			contactState = CONTACTSTATE_OPENED;
			contactEvent = CONTACTEVENT_OPENED;
		}
		break;
		default:
		// UNKNOWN
		if( isOpen )
		{   state = OPENED;
			contactState = CONTACTSTATE_OPENED;
		}
		else
		{   state = CLOSED;
			contactState = CONTACTSTATE_CLOSED;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedSignaling(void)
{
	// Ereignis-Signalausgabe
	UInt08 index = 255;
	switch (contactEvent)
	{   case CONTACTEVENT_CLOSED: index = 0; break;
		case CONTACTEVENT_OPENED: index = 1; break;
		default:                  ;
	}
	if (index != 255)
	{   ToneEventSignal toneSignal;
		STORING_Load(STO_ContactEventConfiguration[index].toneSignal, &toneSignal);
		TONESIGNAL_Play(toneSignal);
		LightEventSignal lightSignal;
		STORING_Load(STO_ContactEventConfiguration[index].lightSignal, &lightSignal);
		UInt08 brightness;
		STORING_Load(STO_ContactEventConfiguration[index].brightness, &brightness);
		LIGHTSIGNAL_Play(lightSignal, brightness);
	}
	// Zustands-Signalausgabe
	{	ToneStateSignal  toneSignal  = TONESTATESIGNAL_OFF;
		LightStateSignal lightSignal = LIGHTSTATESIGNAL_OFF;
		UInt08           brightness  = 10;
		UInt08           index       = 255;
		switch (contactState)
		{   case  CONTACTSTATE_CLOSED: index = 0; break;
			case  CONTACTSTATE_OPENED: index = 1; break;
			default:                   ;
		}
		if (index != 255)
		{   ContactStateSignal tmpSignal;
			STORING_Load(STO_ContactStateSignal[index], &tmpSignal);
			if (toneSignal  == TONESTATESIGNAL_OFF)  toneSignal  = tmpSignal.toneSignal;
			if (lightSignal == LIGHTSTATESIGNAL_OFF) lightSignal = tmpSignal.lightSignal;
			brightness = tmpSignal.brightness;
		}
		if (TONESIGNAL_Current() != toneSignal)
		{	TONESIGNAL_Start(toneSignal);
		}
		if (LIGHTSIGNAL_Current() != lightSignal || LIGHTSIGNAL_Brightness() != brightness)
		{	LIGHTSIGNAL_Start(lightSignal, brightness);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedPushMessaging(void)
{
	const PushMessage* eepPushMessage = NULL;
	switch (contactEvent)
	{   case CONTACTEVENT_CLOSED:
		eepPushMessage = STORING_AddressOf(STO_ContactEventConfiguration[0].pushMessage);
		break;
		case CONTACTEVENT_OPENED:
		eepPushMessage = STORING_AddressOf(STO_ContactEventConfiguration[1].pushMessage);
		break;
		default:
		// CONTACTEVENT_NONE
		;
	}
	if (eepPushMessage != NULL)
	{   PushMessage pushMessage;
		PUSHMESSAGE_EEPLoad(eepPushMessage, &pushMessage);
		if (PUSHMESSAGE_IsValid(&pushMessage))
		{   currentPushMessage = eepPushMessage;
			Cy4MessageSetup messageSetup = SetupContactPushMessage;
			if (!CY4NET_IsPending(messageSetup, NULL)) CY4NET_Post(messageSetup, NULL);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedTextFeeds(void)
{
	const TextFeed* eepTextFeed = NULL;
	switch (contactEvent)
	{   case CONTACTEVENT_CLOSED:
		eepTextFeed = STORING_AddressOf(STO_ContactEventConfiguration[0].textFeed);
		break;
		case CONTACTEVENT_OPENED:
		eepTextFeed = STORING_AddressOf(STO_ContactEventConfiguration[1].textFeed);
		break;
		default:
		// CONTACTEVENT_NONE
		;
	}
	if (eepTextFeed != NULL)
	{   TextFeed textFeed;
		TEXTFEED_EEPLoad(eepTextFeed, &textFeed);
		if (!TEXTFEED_IsEmpty(&textFeed))
		{   currentTextFeed = eepTextFeed;
			Cy4MessageSetup messageSetup = SetupContactTextFeed;
			if (!CY4NET_IsPending(messageSetup, NULL)) CY4NET_Post(messageSetup, NULL);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedSystemLedSignaling(void)
{
	SystemLedSignal signal;
	STORING_Load(STO_SystemLedSignal, &signal);
	bool ledOn = false;
	UInt16 ticks = BOARD_MSecTicks();
	bool heartbeat = (bool)((ticks & 0x380)==0x80 || (ticks & 0x380)==0x180);
	bool busAct    = BOARD_BusActivity();
	switch (signal)
	{	case SYSTEM_LED_SIGNAL_ON:			      ledOn = true;                break;
		case SYSTEM_LED_SIGNAL_HEARTBEAT:	      ledOn = heartbeat;           break;
		case SYSTEM_LED_SIGNAL_BUS_ACT:		      ledOn = busAct;              break;
		case SYSTEM_LED_SIGNAL_HEARTBEAT_BUS_ACT: ledOn = heartbeat || busAct; break;
		default:			                      ;
	}
	BOARD_SetSystemLed(ledOn);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedTimerAutomatic(void)
{
	DateTime dateTime;
	BOARD_Clock(&dateTime);
	if (!DATETIME_IsValid(&dateTime))
	{   BOARD_SetSwitch(false);
		return;
	}
	Time currentTime = dateTime.time;
	bool on = false;
	for (UInt08 n=0; n<3; n++)
	{   AutoTimer autoTimer;
		STORING_Load(STO_AutoTimer[n], &autoTimer);
		if (!autoTimer.enabled ||
		!TIME_IsValid(&autoTimer.startTime) ||
		!TIME_IsValid(&autoTimer.endTime))
		{   continue;
		}
		if (TIME_Compare(&autoTimer.startTime, &autoTimer.endTime) > 0)
		{   // Bsp: Von 23:10 bis 07:15
			on = TIME_Compare(&currentTime, &autoTimer.startTime) >= 0 ||
			TIME_Compare(&currentTime, &autoTimer.endTime) < 0;

		}
		else
		{   // Bsp: Von 10:30 bis 15:10
			on = TIME_Compare(&currentTime, &autoTimer.startTime) >= 0 &&
			TIME_Compare(&currentTime, &autoTimer.endTime) < 0;
		}
		if (on)
		{   DTWeekday weekday = DATE_WeekDay(&dateTime.date);
			if (weekday != DT_NOD)
			{   on = (bool)autoTimer.weekDay[(UInt08)weekday-1];
			}
		}
		if (on) break;  // "Ein" dominiert "Aus"
	}
	BOARD_SetSwitch(on);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedBrightnessAutomatic(void)
{
	if (ambientBrightness == CFG_INVALID_BRIGHTNESS)
	{   BOARD_SetSwitch(false);
		return;
	}
	// (1) Aktuellen Helligkeitszustand ermitteln
	enum
	{   LIGHT_STAGE,    // Zustand Hell
		INTERSTAGE,     // Zwischenzustand
		DARK_STAGE      // Zustand Dunkel
	} stage = INTERSTAGE;
	UInt08 lightnessThreshold;
	STORING_Load(STO_LightnessThreshold, &lightnessThreshold);
	UInt08 darknessThreshold;
	STORING_Load(STO_DarknessThreshold, &darknessThreshold);
	if      (ambientBrightness > lightnessThreshold) stage = LIGHT_STAGE;
	else if (ambientBrightness < darknessThreshold)  stage = DARK_STAGE;
	else                                             stage = INTERSTAGE;
	// (2) Schaltzustand bestimmen
	BrightnessMode mode;
	STORING_Load(STO_BrightnessMode, &mode);
	bool setSwitch = BOARD_Switch();
	switch (mode)
	{   case BRIGHTNESS_MODE_LIGHT_OFF_DARK_ON:
		if      (stage == LIGHT_STAGE) setSwitch = false;
		else if (stage == DARK_STAGE)  setSwitch = true;
		break;
		case BRIGHTNESS_MODE_LIGHT_ON_DARK_OFF:
		if      (stage == LIGHT_STAGE) setSwitch = true;
		else if (stage == DARK_STAGE)  setSwitch = false;
		break;
		default:
		;
	}
	// (3) Automatische Abschaltung
	UInt08 time;
	STORING_Load(STO_BrightnessAutoOffTime, &time);
	if (time > 0)
	{   static Timer autoOffTimer= TIMER_SECONDS(0);
		static UInt08 restarts       =  0;
		if (setSwitch)
		{   if (TIMER_Duration(&autoOffTimer) == 0)
			{   TIMER_Start(&autoOffTimer, 600); // Zeiteinheit: 10 Minuten
				restarts = time-1;
			}
			else if (TIMER_HasExpired(&autoOffTimer))
			{   if (restarts == 0)
				{   setSwitch = false;
				}
				else
				{   restarts--;
					TIMER_Restart(&autoOffTimer);
				}
			}
		}
		else
		{   TIMER_Start(&autoOffTimer, 0);
		}
	}
	// (4) Schaltausgang schalten
	BOARD_SetSwitch(setSwitch);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedOperationMode(void)
{
	if (IsExtraOn()) return;
	SwitchMode mode;
	STORING_Load(STO_PowerSwitchMode, &mode);
	switch (mode)
	{   case SWITCH_MODE_ON:         BOARD_SetSwitch(true);   break;
		case SWITCH_MODE_OFF:        BOARD_SetSwitch(false);  break;
		case SWITCH_MODE_TIMER:      ProceedTimerAutomatic();      break;
		case SWITCH_MODE_BRIGHTNESS: ProceedBrightnessAutomatic(); break;
		default:
		;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedExtraOn(void)
{
	if (IsExtraOn()) BOARD_SetSwitch(true);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Startup(void)
{
    static const Cy4ServiceItem  servicesTable[] PROGMEM =
    {   //|                   service                    |           handler           |
        //+----------------------------------------------+-----------------------------+		
		{ CY4SERVICE(0x100100,  1, CY4SERVICE_READ_WRITE), OnSwitchMode                },
		{ CY4SERVICE(0x100110,  1, CY4SERVICE_READ_ONLY ), OnSwitchState               },
		{ CY4SERVICE(0x100116,  1, CY4SERVICE_WRITE_ONLY), OnExtraOn                   },
		{ CY4SERVICE(0x10011A,  1, CY4SERVICE_READ_WRITE), OnExtraOnDuration           },
		{ CY4SERVICE(0x100120, 12, CY4SERVICE_READ_WRITE), OnTimerModeParameter        },
		{ CY4SERVICE(0x100130, 12, CY4SERVICE_READ_WRITE), OnTimerModeParameter        },
		{ CY4SERVICE(0x100140, 12, CY4SERVICE_READ_WRITE), OnTimerModeParameter        },
		{ CY4SERVICE(0x100180,  4, CY4SERVICE_READ_WRITE), OnBrightnessModeParameter   },
			
		{ CY4SERVICE(0x100410,  7, CY4SERVICE_READ_WRITE), OnDateTime                  },
		{ CY4SERVICE(0x100510,  1, CY4SERVICE_READ_WRITE), OnAmbientBrightness         },
		{ CY4SERVICE(0x100610,  2, CY4SERVICE_READ_WRITE), OnAcceptBroardcasts         },
		{ CY4SERVICE(0x100630, 10, CY4SERVICE_READ_WRITE), OnServerLink                },		

        { CY4SERVICE(0x130000,  1, CY4SERVICE_READ_ONLY) , OnContactState              },
        { CY4SERVICE(0x130010,  6, CY4SERVICE_READ_WRITE), OnContactStateSignal        },
        { CY4SERVICE(0x130100, 61, CY4SERVICE_READ_WRITE), OnContactEventConfiguration },
        { CY4SERVICE(0x130180, 61, CY4SERVICE_READ_WRITE), OnContactEventConfiguration },        		
		{ CY4SERVICE(0x140000,  2, CY4SERVICE_WRITE_ONLY), OnContactEventTest          },		
			
		{ CY4SERVICE(0xE00020,  7, CY4SERVICE_WRITE_ONLY), OnDateTime                  },
		{ CY4SERVICE(0xE00060,  1, CY4SERVICE_WRITE_ONLY), OnAmbientBrightness         },
				        
		{ CY4SERVICE(0xFFF020,  1, CY4SERVICE_READ_WRITE), OnSystemLedSignal           },			
		{ CY4SERVICE(0xFFF030,  2, CY4SERVICE_READ_WRITE), OnBusActivityOnExternalLed  },
        { CY4SERVICE(0xFFFF00, 61, CY4SERVICE_READ_WRITE), OnDeviceLabel               },
        { CY4SERVICE(0xFFFF40,  1, CY4SERVICE_WRITE_ONLY), OnFactorySettings           },
        { CY4SERVICE(0xFFFFF0,  7, CY4SERVICE_READ_ONLY) , OnCallTypePlate             },
        { CY4SERVICE(0xFFFFF8,  1, CY4SERVICE_WRITE_ONLY), OnEnterApplication          },
        { CY4SERVICE(0xFFFFFA,  1, CY4SERVICE_WRITE_ONLY), OnEnterProgramLoader        },
        { CY4SERVICE(0xFFFFFC,  2, CY4SERVICE_WRITE_ONLY), OnChangeCy4Address          }
    };
    CY4NET_AssignServices(servicesTable);
    Cy4Address address = STORING_LoadCy4SelfAdr();
    CY4NET_SetAddress(address);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Proceed(void)
{
    ProceedContacting();
    ProceedSignaling();
    ProceedPushMessaging();
    ProceedTextFeeds();
	ProccedServerRequests();
	ProceedSystemLedSignaling();	
	ProceedOperationMode();
	ProceedExtraOn();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Shutdown(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
