﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Boardmodul: Bereitstellung gerätespezifischer Funktionen und Mechanismen            //
//  _________________________________  													//
//																						//
//  Copyright (c) 2018 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include "Conf.h"
#include "AVRChip.h"
#include "Timer.h"
#include "Board.h"
#include "Storing.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Definitionen  ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define Sw_Set  C,1
#define Ct_In   C,2
#define SysLED  D,3

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                        // Arbeitsanweisungen für das Board-Modul:
{   DO_STARTUP,                     // · Starte die Anwendung neu
    DO_ENTER_PGMLDR,                // · Starte den Programmlader
    DO_PROCEED,                     // · Führe die Arbeit fort
    DO_SHUTDOWN                     // · Schalte das Board ab
} DoNext;

typedef struct                      // Bus-Übertragungsstatus:
{   bool   pending : 1;             // · Übertragung im Gange
    bool   echo    : 1;             // · Echo erhalten
    bool   error   : 1;             // · Übertragung Fehlgeschlagen (Kollision)
    UInt08 byteGap : 5;             // · Vergangene Zeit seit dem letzten Byteempfang [ms]
} BusStatus;

typedef struct                      // Element der Bus-Empfangsschlange:
{   bool   timeout : 1;             // · Zeitüberschreitung bei <byteGap> aufgetreten
    UInt08 padding : 7;             // · Füllbits (ohne Bedeutung)
    UInt08 data    : 8;             // · Datenbyte
} BusRxItem;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static volatile UInt16     msecTicks                             = 0;
static volatile UInt16     secTicks                              = 0;
static volatile UInt08     busTxQueue[CFG_BUS_TX_QUEUE_CAPACITY] = { 0 };
static volatile UInt08     busTxQueueHead                        = 0;
static volatile UInt08     busTxQueueLength                      = 0;
static volatile UInt08     busLoopbackByte                       = 0;
static volatile BusStatus  busStatus                             = { false, false, false, 0};
static volatile BusRxItem  busRxQueue[CFG_BUS_RX_QUEUE_CAPACITY] = { { 0, false, false } };
static volatile UInt08     busRxQueueHead                        = 0;
static volatile UInt08     busRxQueueTail                        = 0;
static          DoNext     doNext                                = DO_STARTUP;
static          UInt08     busActivityIndicator                  = 0;
static          UInt08     externalLedBrightness                 = 0;
static          UInt08     toneSustain                           = 0;
static          DateTime   systemDateTime                        = DATETIME_INVALID;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static UInt08
LedGammaCorrection(UInt08 percent)
{
	// 0 .. 100 [%] -> 0 .. 255 [pwm] mit Gammakorrektur 0.5
	if( percent < 100 ) return (UInt08)(((SInt16)percent*percent)/39);
    else				return 255;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedTonePlaying(void)
{
    static Timer timer = TIMER_MILLISECONDS(10);

    if( toneSustain > 0 )
    {   if( TIMER_RestartIfExpired(&timer) )
        {   toneSustain--;
            if( toneSustain == 0 )
            {   TIMER_Stop(&timer);
                BOARD_PlayTone(NOTE_OFF, 0);
            }
		}
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Reset(): Führt einen physikalischen Reset herbei indem der Watchdog ausgelöst wird. //
//  -------  Reset() kehrt nie zum Aufrufer zurück.                                     //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((noreturn)) static void
Reset(void)
{
    DisableInterrupts();            // Interrupts sperren
    wdt_enable(WDTO_15MS);          // Watchdog-Intervall auf kürzeste Zeit setzen
    __asm__ volatile
    (   "ldi r30, lo8 (%0) \n"      // Nach einem Neustart des Systems wird als
	    "ldi r31, hi8 (%0) \n"      // erstes das Programmladermodul ausgeführt.
	    "ldi r24, %1       \n"      // Der Programmlader nutzt zur Steuerung seines
	    "ldi r25, hi8 (%2) \n"      // Ablaufs Variablen, die in der "Noinit"-Sektion
	    "0:                \n"      // angesiedelt sind, d.h. sie werden im Rahmen
	    "st  Z+,  r24      \n"      // eines Neustartes nicht initialisiert.
	    "cpi r30, lo8 (%2) \n"      // Um nun zu vermeiden, dass übrig gebliebene
	    "cpc r31, r25      \n"      // Werte des SRAMs unerwünschte Effekte im
	    "brlo 0b"                   // Programmlader hervorrufen, wird vor dem
	        :                       // Reset das komplette SRAM mit Nullen überschrieben.
	        : "i" (RAMSTART),       // Da durch die Nullung Heap wie Stack komplett
              "i" (0),              // zerstört sind, ist diese Aktion ein Punkt-Ohne-
              "i" (RAMEND+1)        // Wiederkehr.
	);
    while( true );                  // Warten, bis Watchdog-Reset eintritt
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
EnterPgmLdr(void)
{
	if (pgm_read_word(FLASHEND - CFG_PGMLDR_SIZE*1024 + 1) == 0x940C)
	{   //GOTO(FLASHEND - CFG_PGMLDR_SIZE*1024 + 1 + 4);
		// Seit dem ATmega328PB ist FLASHEND definiert als "0x7FFFU". Das abschließende
		// "U" funktioniert im Inline-Assembler von GOTO nicht mehr. Daher ...
		#if FLASHEND == 0x7FFFU
		GOTO(0x7FFF - CFG_PGMLDR_SIZE*1024 + 1 + 4);
		#else
		#error malformed FLASHEND
		#endif
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_PreStart(): Diese Funktion wird einmal nach dem Systemstart aufgerufen. Der   //
//  ----------------  Zeitpunkt ihres Aufrufs macht sie zur einer Sonderfunktion, denn  //
//  ihr Aufruf findet in der Start-Sektion "init3", noch lange vor Aufruf von main(),   //
//  statt. Das ist der richtige Zeitpunkt, um z.B. die Ursache des Resets zu ermitteln  //
//  oder statische Variablen auszulesen, deren Werte vor dem Reset platziert wurden     //
//  und den Neustart überlebt haben. Denn zu diesem Zeitpunkt sind die statischen       //
//  Variablen noch nicht initialisiert.                                                 //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((naked, section(".init3"))) void
BOARD_PreStart(void)
{
    MCUSR = 0;
    wdt_disable();
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedSystemTime(void)
{
	if (DATETIME_IsValid(&systemDateTime))
	{   static Timer update = TIMER_MILLISECONDS(1000);
		if (TIMER_RestartIfExpired(&update)) DATETIME_ShiftSecond(&systemDateTime, 1);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Interruptbehandler  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
ISR(TIMER2_COMPA_vect)
{
	// Abweichungskompensation              // Quarztaktbedingt findet der IRQ-Aufruf
    {   static UInt08 counter = 0;          // bei einem Zeitgeber-Zählerwert von 42
        if( counter == 4 )                  // nicht exakt jede Millisekunde statt,
        {   counter = 0;                    // sondern etwas schneller:
            OCR2A = F_CPU/(256*1000UL);     //
        }                                   //      43 ÷ 43.2 = 0.9954 ms
        else                                //
        {   counter++;                      // Zwecks Ausgleichs dieser Ungenauigkeit
            OCR2A = F_CPU/(256*1000UL)-1;   // wird bei jedem fünften Aufruf, der
        }                                   // Überlauf um einen Zählerwert verlängert:
    }                                       // 4·(42÷42.2)ms + (
    // Ticker inkrementieren
    {   msecTicks++;
        static UInt16 counter = 0;
        counter++;
        if( counter == 1000 )
        {   counter =  0;
            secTicks++;
        }
    }
    // Bus Byte-Zeitlücke
    if( busStatus.byteGap <= CFG_BUS_INTERBYTE_TIMEOUT )
    {   busStatus.byteGap++;
        if( busStatus.byteGap > CFG_BUS_INTERBYTE_TIMEOUT )
        {   busRxQueue[busRxQueueTail].timeout = true;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART0_TX_vect)
{                                   // IRQ Abfolge sowie die Signale <echo> und <pending>:
	if (!busStatus.echo)            //
	{   busStatus.error = true;     // Start Tx:    |1|
		busTxQueueLength = 0;       //               :
	}                               // Tx-IRQ:  ···--:-----|2|----|3|-···-|n|----| |-···
	busStatus.echo = false;         // Rx-IRQ:  ···--:-|1|--:-|2|--:--···-- -|n|--:--···
	UInt08 head = busTxQueueHead;   //               :  :___:  :___:          :___:
	if (busTxQueueLength == 0)      // echo:    ...__:__|   |__|   |____...___|   |__...
	{   busStatus.pending = false;  //               :____________________________:
	}                               // pending: ...__|                            |__...
	else
	{   busLoopbackByte = busTxQueue[head];
		UDR0            = busTxQueue[head];
		head++;
		if (head == CARDINALITY(busTxQueue)) head = 0;
		busTxQueueLength--;
	}
	busTxQueueHead = head;
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART0_RX_vect)
{
	UInt08 byte = UDR0;
	busStatus.byteGap = 0;   // Byte empfangen, Zeitzähler zurücksetzen
	if (busStatus.pending && byte == busLoopbackByte) busStatus.echo = true; // Echo
	// Byte in Empfangsschlange stellen
	UInt08 tail = busRxQueueTail;
	busRxQueue[tail].data = byte;
	tail++;
	if (tail == CARDINALITY(busRxQueue)) tail = 0;
	if (tail == busRxQueueHead) busRxQueue[busRxQueueTail].timeout = true; // Überlauf
	else                        busRxQueueTail = tail;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Funktionen ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Reset(): Initiiert einen Systemneustart. Die Funktion löst nicht unmittelbar  //
//  -------------  einen Reset aus, sondern führt zu einem Ausstieg aus der Haupt-      //
//  schleife von Main(). Die danach sukzessiv aufgerufenen "Shutdown"-Funktionen aller  //
//  betroffenen Module bewirken den kontrollierten Abschluss aller Aktivitäten, bevor   //
//  die letzte aufgerufene "Shutdown"-Funktion des Board-Moduls dann den physikali-     //
//  schen Reset auslöst.                                                                //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Reset(void)
{
    doNext = DO_STARTUP;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EnterPgmLdr(): Ähnlich wie BOARD_Reset() initiiert BOARD_EnterPgmLdr() einen  //
//  -------------------  Systemneustart. Im Unterschied zu BOARD_Reset(), wird hier     //
//  jedoch nicht die Anwendung neu gestartet, sondern das System bootet in das Pro-     //
//  grammladermodul.                                                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EnterPgmLdr(void)
{
    doNext = DO_ENTER_PGMLDR;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_MSecTicks(): Die Funktion liefert den Stand eines monotonen Millisekunden-    //
//  -----------------  Zeitzählers. Der Zähler wird interruptgetrieben inkrementiert    //
//  und hat eine Breite von 16-Bit. Kommt es nach 65535 ms zum Überlauf, beginnt der    //
//  Zeitzähler wieder bei 0.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Millisekunden-Zeitzählers.                             //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_MSecTicks(void)
{
    UInt16 ticks = 0;

    Atomic { ticks = msecTicks; }
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SecTicks(): Die Funktion liefert den Stand eines monotonen Sekunden-Zeit-     //
//  ----------------  zählers. Der Zähler wird interruptgetrieben inkrementiert und hat //
//  eine Breite von 16-Bit. Kommt es nach 65535 s (knapp 18¼h) zum Überlauf, beginnt    //
//  der Zeitzähler wieder bei 0.                                                        //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Sekunden-Zeitzählers.                                  //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_SecTicks(void)
{
    UInt16 ticks = 0;

    Atomic { ticks = secTicks; }
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EEPLoad(): Kopiert <count> Bytes aus dem EEPROM-Speicher ins RAM.             //
//  ---------------                                                                     //
//                                                                                      //
//  Eingabe: · eepAddress:  Adresse im EEPROM-Speicher.                                 //
//           · ramAddress:  Adresse im Ram.                                             //
//           · count:       Anzahl zu kopierender Bytes.                                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPLoad(const void* eepAddress, void* ramAddress, UInt16 count)
{
    eeprom_read_block(ramAddress, (void*)eepAddress, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EEPSave(): Kopiert <count> Bytes aus dem RAM in den EEPROM-Speicher. Da       //
//  ---------------  EEPROM-Speicher lt. Datenblatt nur ca. 100000 Lösch/Schreibzyklen  //
//  pro Zelle verträgt, wird vor für jedes Bytes vor dem Speichervorgang geprüft, ob    //
//  sich der zu schreibende Wert sich vom Inhalt der EEP-Speicherzelle unterscheidet.   //
//  Erst danach wird nötigenfalls ein Lösch- und Schreibvorgang durchgeführt.           //
//  Hinweis: Man beachte, dass das Beschreiben einer EEPROM-Speicherzelle ca. 8 ms in   //
//           Anspruch nimmt und damit verhältnismäßig viel Zeit verschlingt.            //
//                                                                                      //
//  Eingabe: · ramAddress:  Adresse im Ram.                                             //
//           · eepAddress:  Adresse im EEPROM-Speicher.                                 //
//           · count:       Anzahl zu kopierender Bytes.                                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPSave(const void* ramAddress, void* eepAddress, UInt16 count)
{
    eeprom_update_block(ramAddress, (void*)eepAddress, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsIdle(): Prüft, ob der Bus momentan frei ist. Der Bus gilt als frei, wenn //
//  -----------------  eine Kommunikationspause von mehr als CFG_BUS_INTERBYTE_TIMEOUT  //
//  Millisekunden eintritt.                                                             //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Bus frei ist.                                             //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsIdle(void)
{
    bool  idle = false;
    Atomic
    {   idle = !busStatus.pending && busStatus.byteGap > CFG_BUS_INTERBYTE_TIMEOUT;
    }
    return idle;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusTransmit(): Sendet das Datenbyte in <byte> auf den Bus.                    //
//  -------------------  Die Sendeeinheit ist durch folgenden Zustandsautomaten         //
//  realisiert, Ausgangszustand ist IDLE:                                               //
//                                                                                      //
//                                                                                      //
//     +----------+ BOARD_BusTransmit()    +------------------+ BOARD_BusTransmit()     //
//     |//////////+----------------------->|//////////////////+--------------------+    //
//     |// IDLE //|              COMPLETED |// TRANSMITTING //|                    |    //
//     |//////////|<-----------------------+//////////////////|<-------------------+    //
//     +----------+                        +--------+---------+                         //
//          /|\                                     | COLLISION                         //
//           |                                      |                                   //
//           |                                     \|/                                  //
//           |                               +--------------+                           //
//           | BOARD_BusTakeFailure()        |//////////////|                           //
//           +-------------------------------+// FAIILURE //|                           //
//                                           |//////////////|                           //
//                                           +--------------+                           //
//                                                                                      //
//  Mit Aufruf dieser Funktion wechselt der Automat IDLE nach TRANSMITTING und beginnt  //
//  mit der Übertragung des angegebenen Datenbytes. Ist bereits eine Übertragung im     //
//  Gange (Zustand TRANSMITTING), wird bei Aufruf der Funktion das zu übertragende      //
//  Datenbyte für die spätere Übertragung in die Sende-Warteschlange gestellt. Nach     //
//  Aussendung aller zu übertragenden Datenbytes (Ereignis COMPLETED) kehrt der Automat //
//  wieder in den Zustand IDLE zurück.                                                  //
//  Das Halbduplexverfahren des seriellen Busses bedingt, dass jedes ausgesendete Byte  //
//  unmittelbar als Echo wieder in den Empfangspuffer zurückwandert. Diese Rücküber-    //
//  tragung wird genutzt, um Kollisionen aufzudecken. Unterscheiden sind ausgesendetes  //
//  Byte und Echo, löst das das Ereignis COLLISION aus und damit den Wechsel in den     //
//  Zustand FAIILURE. Die Übertragung wird in dem Fall nicht mehr fortgesetzt und die   //
//  Sende-Warteschlange geleert. Gegebenenfalls noch ausstehende Bytes werden also      //
//  nicht mehr übertragen. Im Zustand FAIILURE bleiben Aufrufe von BOARD_BusTransmit()  //
//  ohne Wirkung. Um neue Datenbytes senden zu können, muss der Automat erst durch      //
//  Aufruf von BOARD_BusTakeFailure() wieder in den Zustand IDLE versetzt werden.       //
//                                                                                      //
//  Eingabe: · byte: Das zu übertragende Datenbyte.                                     //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusTransmit(UInt08 byte)
{
    // Ist die Übertragungsschlange voll, warten, bis wieder Platz ist
    bool  queueFull = false;
    do
    {   Atomic
        {   queueFull = busTxQueueLength == CARDINALITY(busTxQueue);
	    }
    } while( queueFull );
    // Byte senden
    Atomic
    {   if( busStatus.error ) return; // Zustand FAIILURE: Nichts tun
        if( busStatus.pending )
        {   // Zustand TRANSMITTING: Byte für Übertragung aufreihen
            UInt08 tail = busTxQueueHead + busTxQueueLength;
            if( tail >= CARDINALITY(busTxQueue) ) tail -= CARDINALITY(busTxQueue);
            busTxQueue[tail] = byte;
            busTxQueueLength++;
        }
        else
        {   // Zustand IDLE: Byte unmittelbar senden
            busStatus.pending = true;
            busStatus.echo    = false;
            busLoopbackByte   = byte;
            UDR0              = byte;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsTransmitting(): Prüft, ob sich der Automat (siehe BOARD_BusTransmit())   //
//  -------------------------  im Zustand TRANSMITTING befindet. Das ist solange der    //
//  Fall, wie die Sendeeinheit mit der Übertragung von Datenbytes beschäftigt ist.      //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Sendevorgang noch im Gange ist.                           //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsTransmitting(void)
{
    bool  pending = false;
    Atomic
    {   pending = busStatus.pending;
    }
    return pending;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusTakeFailure(): Prüft, ob sich der Automat (siehe BOARD_BusTransmit()) im   //
//  ----------------------  Zustand FAILURE befindet und löst gegebenenfalls einen      //
//  Wechsel in den Zustand IDLE aus. Übertragungsfehler entstehen in der Regel durch    //
//  Kollisionen auf dem Bus.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn ein Übertragungsfehler vorlag.                                //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusTakeFailure(void)
{
    bool  error = false;
    Atomic
    {   error = busStatus.error;
        busStatus.error = false;
    }
    return error;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusReceive(): Steht bei Aufruf der Funktion ein empfangenes Datenbyte zur     //
//  ------------------  Abholung bereit, wird es aus der Empfangsschlange entfernt und  //
//  in <byte> übergeben. BOARD_BusReceive() kehrt so lange mit 'true' zurück, wie       //
//  Datenbytes zur Abholung bereitstehen.                                               //
//  Man beachte: Der Rückgabewert von BOARD_BusReceive() sagt nichts darüber aus, ob    //
//  eine Empfangssequenz noch im Gange ist oder nicht. Dazu siehe BOARD_BusIs-          //
//  Receiving().                                                                        //
//                                                                                      //
//  Eingabe: · byte: Adresse des Bytes, das das nächste zur Abholung bereitstehende     //
//                   Datenbyte aufnimmt.                                                //
//  Ausgabe: 'true', wenn in <byte> ein Datenbyte abgelegt wurde.                       //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusReceive(UInt08* byte)
{
    bool  hasData = false;
    Atomic
    {   UInt08 head = busRxQueueHead;
        if( !busRxQueue[head].timeout )
        {   // Keine Zeitüberschreitungs-Markierung gesetzt
            if( head != busRxQueueTail )
            {   // Datenbyte vorhanden
                *byte = busRxQueue[head].data;
                head++;
                if( head == CARDINALITY(busRxQueue) ) head = 0;
                busRxQueueHead = head;
                hasData = true;
            }
        }
    }
    return hasData;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsReceiving(): Prüft, ob die aktuelle Empfangssequenz noch im Gange ist.   //
//  ----------------------                                                              //
//  Eine Gruppe empfangener Datenbytes gilt als zusammenhängene Empfangssequenz, wenn   //
//  die Zeitspanne zwischen dem Empfang sukzessiver Datenbytes einen bestimmten Zeit-   //
//  rahmen nicht überschreitet. Mit anderen Worten: Liegen die Zeitpunkte zweier auf-   //
//  einanderfolgend empfangener Datenbytes länger als CFG_BUS_INTERBYTE_TIMEOUT Milli-  //
//  sekunden auseinander (timeout), gilt die Empfangssequenz als unterbrochen.          //
//                                                                                      //
//    Empfang: ···-|A|-|B|-|C|---------------×---|D|----|E|---------------×--|F|-···    //
//                           :               :            :               :             //
//                           |--- timeout ---|            |--- timeout ---|             //
//                                                                                      //
//	  Empfangssequenzen: ABC, DE, F... 	                                                //
//                                                                                      //
//  Die Unterbrechungen werden mit in der Empfangsschlange vermerkt, so dass beim Aus-  //
//  lesen der empfangenen Bytes mittels BOARD_BusReceive() die Empfangssequenzen zeit-  //
//  unabhängig separiert werden. Solange BOARD_BusIsReceiving() mit 'true' zurückkehrt, //
//  ist die Empfangssequenz noch im Gange.                                   			//
//                                                                                      //
//      Beispiel: Empfang von Byte-Sequenzen:                                           //
//                                                                                      //
//       UInt08 sequence[SEQUENCE_CAPACITY];   // SEQUENCE_CAPACITY: 1 .. 255           //
//                                                                                      //
//       UInt08                                                                         //
//       TakeNextSequence(void)                                                         //
//       {                                                                              //
//           static UInt08 index = 0;                                                   //
//                                                                                      //
//           UInt08 length = 0;                                                         //
//           if( BOARD_BusIsReceiving() )                                               //
//           {   // Empfang einer Sequenz im Gange                                      //
//               UInt08 byte;                                                           //
//               while( BOARD_BusReceive(&byte) ) // Datenbytes abholen und aufsammeln  //
//               {   if( index < SEQUENCE_CAPACITY )                                    //
//                   {   sequence[index] = byte;                                        //
//                       index++;                                                       //
//                   }                                                                  //
//                   else                                                               //
//                   {   // Überlauf ...                                                //
//                   }                                                                  //
//               }                                                                      //
//           }                                                                          //
//           else                                                                       //
//           {    // Empfang einer Sequenz nicht (mehr) im Gange                        //
//                length = index;                                                       //
//                index  = 0;                                                           //
//           }                                                                          //
//           return length;   // Rückgabewert: Aktuelle Sequenzlänge                    //
//       }                                                                              //
//                                                                                      //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', solange die Empfangssequenz im Gange ist.                          //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsReceiving(void)
{
    bool  timeout = false;
    Atomic
    {   UInt08 head = busRxQueueHead;
        timeout = busRxQueue[head].timeout;
        if( head != busRxQueueTail ) busRxQueue[head].timeout = false;
    }
    return !timeout;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusOnActivity(): Rückruffunktion der Bus-Übertragungseinheit. Die Funktion    //
//  ---------------------  wird aufgerufen, wenn mit dem Gerät über den Bus             //
//  kommuniziert wird.                                                                  //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusOnActivity(void)
{
    busActivityIndicator = CFG_BUS_ACT_INDICATOR_DURATION;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusActivity(): Der Bus-Aktivitätsindikator zeigt an, ob aktuell mit dem       //
//  -------------------  Gerät kommuniziert wird. Diese Funktion liefert den Zustand    //
//  des Indikators. Nach Abschluss der Kommunikation bleibt der Indikator noch 'CFG_-   //
//  BUS_ACT_INDICATOR_DURATION' Millisekunden aktiviert.                                //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusActivity(void)
{
	return busActivityIndicator != 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SetClock(): Stellt sie Systemuhr mit der angegebenen Datumzeit. Ist die       //
//  ----------------  übergebene Datumzeit ungültig, wird die Systemuhr angehalten.     //
//                                                                                      //
//  Eingabe: · dateTime: Datumzeit, mit der die Systemuhr gestellt wird.                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetClock(const DateTime* dateTime)
{
	systemDateTime = *dateTime;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Clock(): Liest die Systemuhr aus. In dem Fall, dass die Systemuhr nicht läuft //
//  -------------  ist die zurückgelieferte Datumzeit ungültig.                         //
//                                                                                      //
//  Eingabe: · dateTime: Referenz, die die Zeitinformation aufnimmt.                    //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Clock(DateTime* dateTime)
{
	*dateTime = systemDateTime;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_PlayTone(): Spielt den angegebenen Ton.                                       //
//  ----------------                                                                    //
//                                                                                      //
//  Eingabe: · note:    Die abzuspielende Tonhöhe gemäß Aufzählungstyp 'Note'.          //
//           · sustain: Haltedauer: 0 .. 250 [1÷100s], 0 = Dauerton                     //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_PlayTone(Note note, UInt08 sustain)
{
	static const UInt08  lookupTable[][2] PROGMEM =  // Tonfrequenz-Registerwerte LUT
	{   //  N ,  OCR      Index  Enum      Ton = Frequenz [Hz] | µC-Freq [Hz] (Abweichung)
		//-----------------------------------------------------+--------------------------
		{  0 , 0x00 }, //  [ 0]  NOTE_OFF  --                  |
		{  4 , 0xA4 }, //  [ 1]  NOTE_C3   C3  =   130.813     |    130.909     ( 0.07%)
		{  4 , 0x9B }, //  [ 2]  NOTE_C3f  C3# =   138.591     |    138.462     (-0.09%)
		{  4 , 0x92 }, //  [ 3]  NOTE_D3   D3  =   146.832     |    146.939     ( 0.07%)
		{  4 , 0x8A }, //  [ 4]  NOTE_D3f  D3# =   155.563     |    155.676     ( 0.07%)
		{  4 , 0x82 }, //  [ 5]  NOTE_E3   E3  =   164.814     |    164.571     (-0.15%)
		{  4 , 0x7B }, //  [ 6]  NOTE_F3   F3  =   174.614     |    174.545     (-0.04%)
		{  4 , 0x74 }, //  [ 7]  NOTE_F3f  F3# =   184.997     |    184.615     (-0.21%)
		{  4 , 0x6D }, //  [ 8]  NOTE_G3   G3  =   195.998     |    195.918     (-0.04%)
		{  4 , 0x67 }, //  [ 9]  NOTE_G3f  G3# =   207.652     |    207.194     (-0.22%)
		{  4 , 0x61 }, //  [10]  NOTE_A3   A3  =   220.000     |    219.847     (-0.07%)
		{  4 , 0x5C }, //  [11]  NOTE_B3   B3  =   233.082     |    232.258     (-0.35%)
		{  4 , 0x56 }, //  [12]  NOTE_H3   H3  =   246.942     |    246.154     (-0.32%)
		{  4 , 0x52 }, //  [13]  NOTE_C4   C4  =   261.626     |    261.818     ( 0.07%)
		{  4 , 0x4D }, //  [14]  NOTE_C4f  C4# =   277.183     |    276.923     (-0.09%)
		{  4 , 0x49 }, //  [15]  NOTE_D4   D4  =   293.665     |    293.878     ( 0.07%)
		{  4 , 0x44 }, //  [16]  NOTE_D4f  D4# =   311.127     |    309.677     (-0.47%)
		{  4 , 0x41 }, //  [17]  NOTE_E4   E4  =   329.628     |    331.034     ( 0.43%)
		{  3 , 0xF6 }, //  [18]  NOTE_F4   F4  =   349.228     |    351.220     ( 0.57%)
		{  3 , 0xE9 }, //  [19]  NOTE_F4f  F4# =   369.994     |    369.231     (-0.21%)
		{  3 , 0xDB }, //  [20]  NOTE_G4   G4  =   391.995     |    394.521     ( 0.64%)
		{  3 , 0xCF }, //  [21]  NOTE_G4f  G4# =   415.305     |    417.391     ( 0.50%)
		{  3 , 0xC3 }, //  [22]  NOTE_A4   A4  =   440.000     |    443.077     ( 0.70%)
		{  3 , 0xB8 }, //  [23]  NOTE_B4   B4  =   466.164     |    464.516     (-0.35%)
		{  3 , 0xAE }, //  [24]  NOTE_H4   H4  =   493.883     |    496.552     ( 0.54%)
		{  3 , 0xA4 }, //  [25]  NOTE_C5   C5  =   523.251     |    523.636     ( 0.07%)
		{  3 , 0x9B }, //  [26]  NOTE_C5f  C5# =   554.365     |    553.846     (-0.09%)
		{  3 , 0x92 }, //  [27]  NOTE_D5   D5  =   587.330     |    587.755     ( 0.07%)
		{  3 , 0x8A }, //  [28]  NOTE_D5f  D5# =   622.254     |    626.087     ( 0.62%)
		{  3 , 0x82 }, //  [29]  NOTE_E5   E5  =   659.255     |    654.545     (-0.71%)
		{  3 , 0x7B }, //  [30]  NOTE_F5   F5  =   698.456     |    702.439     ( 0.57%)
		{  3 , 0x74 }, //  [31]  NOTE_F5f  F5# =   739.989     |    738.462     (-0.21%)
		{  3 , 0x6D }, //  [32]  NOTE_G5   G5  =   783.991     |    778.378     (-0.72%)
		{  3 , 0x67 }, //  [33]  NOTE_G5f  G5# =   830.609     |    822.857     (-0.93%)
		{  3 , 0x61 }, //  [34]  NOTE_A5   A5  =   880.000     |    872.727     (-0.83%)
		{  3 , 0x5C }, //  [35]  NOTE_B5   B5  =   932.328     |    932.794     ( 0.05%)
		{  3 , 0x56 }, //  [36]  NOTE_H5   H5  =   987.767     |    988.841     ( 0.11%)
		{  3 , 0x52 }, //  [37]  NOTE_C6   C6  =  1046.502     |   1047.273     ( 0.07%)
		{  3 , 0x4D }, //  [38]  NOTE_C6f  C6# =  1108.731     |   1107.692     (-0.09%)
		{  3 , 0x49 }, //  [39]  NOTE_D6   D6  =  1174.659     |   1175.510     ( 0.07%)
		{  3 , 0x44 }, //  [40]  NOTE_D6f  D6# =  1244.508     |   1245.405     ( 0.07%)
		{  3 , 0x41 }, //  [41]  NOTE_E6   E6  =  1318.510     |   1316.571     (-0.15%)
		{  3 , 0x3D }, //  [42]  NOTE_F6   F6  =  1396.913     |   1396.364     (-0.04%)
		{  3 , 0x39 }, //  [43]  NOTE_F6f  F6# =  1479.978     |   1476.923     (-0.21%)
		{  3 , 0x36 }, //  [44]  NOTE_G6   G6  =  1567.982     |   1567.347     (-0.04%)
		{  3 , 0x33 }, //  [45]  NOTE_G6f  G6# =  1661.219     |   1657.554     (-0.22%)
		{  3 , 0x30 }, //  [46]  NOTE_A6   A6  =  1760.000     |   1758.779     (-0.07%)
		{  3 , 0x2D }, //  [47]  NOTE_B6   B6  =  1864.655     |   1858.065     (-0.35%)
		{  3 , 0x2B }, //  [48]  NOTE_H6   H6  =  1975.533     |   1969.231     (-0.32%)
		{  3 , 0x28 }, //  [49]  NOTE_C7   C7  =  2093.005     |   2094.545     ( 0.07%)
		{  3 , 0x26 }, //  [50]  NOTE_C7f  C7# =  2217.461     |   2215.385     (-0.09%)
		{  3 , 0x24 }, //  [51]  NOTE_D7   D7  =  2349.318     |   2351.020     ( 0.07%)
		{  3 , 0x22 }, //  [52]  NOTE_D7f  D7# =  2489.016     |   2477.419     (-0.47%)
		{  3 , 0x20 }, //  [53]  NOTE_E7   E7  =  2637.020     |   2648.276     ( 0.43%)
		{  2 , 0xF6 }, //  [54]  NOTE_F7   F7  =  2793.826     |   2809.756     ( 0.57%)
		{  2 , 0xE9 }, //  [55]  NOTE_F7f  F7# =  2959.955     |   2953.846     (-0.21%)
		{  2 , 0xDB }, //  [56]  NOTE_G7   G7  =  3135.963     |   3156.164     ( 0.64%)
		{  2 , 0xCF }, //  [57]  NOTE_G7f  G7# =  3322.438     |   3339.130     ( 0.50%)
		{  2 , 0xC3 }, //  [58]  NOTE_A7   A7  =  3520.000     |   3544.615     ( 0.70%)
		{  2 , 0xB8 }, //  [59]  NOTE_B7   B7  =  3729.310     |   3716.129     (-0.35%)
		{  2 , 0xAE }, //  [60]  NOTE_H7   H7  =  3951.066     |   3972.414     ( 0.54%)
		{  2 , 0xA4 }, //  [61]  NOTE_C8   C8  =  4186.009     |   4189.091     ( 0.07%)
		{  2 , 0x9B }, //  [62]  NOTE_C8f  C8# =  4434.922     |   4430.769     (-0.09%)
		{  2 , 0x92 }, //  [63]  NOTE_D8   D8  =  4698.636     |   4702.041     ( 0.07%)
		{  2 , 0x8A }, //  [64]  NOTE_D8f  D8# =  4978.032     |   5008.696     ( 0.62%)
		{  2 , 0x82 }, //  [65]  NOTE_E8   E8  =  5274.041     |   5236.364     (-0.71%)
		{  2 , 0x7B }, //  [66]  NOTE_F8   F8  =  5587.652     |   5619.512     ( 0.57%)
		{  2 , 0x74 }, //  [67]  NOTE_F8f  F8# =  5919.911     |   5907.692     (-0.21%)
		{  2 , 0x6D }, //  [68]  NOTE_G8   G8  =  6271.927     |   6227.027     (-0.72%)
		{  2 , 0x67 }, //  [69]  NOTE_G8f  G8# =  6644.875     |   6582.857     (-0.93%)
		{  2 , 0x61 }, //  [70]  NOTE_A8   A8  =  7040.000     |   6981.818     (-0.83%)
		{  2 , 0x5C }, //  [71]  NOTE_B8   B8  =  7458.620     |   7462.348     ( 0.05%)
		{  2 , 0x56 }, //  [72]  NOTE_H8   H8  =  7902.133     |   7910.730     ( 0.11%)
		{  2 , 0x52 }  //  [73]  NOTE_C9   C9  =  8372.018     |   8378.182     ( 0.07%)
	};

	toneSustain = sustain;	
    UInt08 ocr = pgm_read_byte(&lookupTable[(UInt08)note][1]);
	UInt08 i   = pgm_read_byte(&lookupTable[(UInt08)note][0]);
	static const UInt08 div[] PROGMEM =
	{   0, 						// [0] Ton aus
		BV(CS10), 				// [1] F_CPU ÷   1
		BV(CS11), 				// [2] F_CPU ÷   8
		BV(CS11) | BV(CS10),   	// [3] F_CPU ÷  64
		BV(CS12)                // [4] F_CPU ÷ 256
	};
	UInt08 cs = pgm_read_byte(&div[i]);    
    TCCR1B = BV(WGM12) | cs;
	if( cs == 0 )
    {   TCCR1A = 0;    
        OCR1A  = 0;
    }
    else
    {   if( TCCR1A == 0 || ocr != OCR1A )
        {   TCCR1A = BV(COM1A0);	// Port-Pin OC1A aktivieren
            OCR1A  = ocr;
            TCNT1  = 0;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_IsSilence(void)
{
	return toneSustain == 0  && TCCR1A == 0 && OCR1A == 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_PinState(): Liefert den Zustand des Kontakt-Pins.  							//
//  ----------------																	//
//                                                                                      //
//  Eingabe: -																			//
//  Ausgabe: 'true'/'false' Pinzustand ist hight/Low.									//
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_PinState(void)
{	
	return (bool)GetPin(Ct_In);	
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SetLed(): Setzt die Helligkeit der externen LED.								//
//  --------------                                                                      //
//                                                                                      //
//  Eingabe: · brightness: Helligkeit der LED, 0 .. 100 %.								//
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetLed(UInt08 brightness)
{
	externalLedBrightness = brightness;
	OCR0A = LedGammaCorrection(brightness);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Led(): Liefert die Helligkeit der externen LED.								//
//  -----------																			//
//                                                                                      //
//  Eingabe: -																			//
//  Ausgabe: Helligkeit, 0 .. 100 %.													//
//////////////////////////////////////////////////////////////////////////////////////////
UInt08
BOARD_Led(void)
{
	return externalLedBrightness;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetSystemLed(bool on)
{
	ChgPin(SysLED, on);
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_SystemLed(void)
{
	return (bool)GetPin(SysLED);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetSwitch(bool on)
{
	ChgPin(Sw_Set, on);
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_Switch(void)
{
	return (bool)GetPin(Sw_Set);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Startup(void)
{
    if( doNext != DO_STARTUP ) Reset();
    DisableInterrupts();
    // Watchdog --------------------------------------------------------------------------
    wdt_enable(WDTO_2S);
    // Ports'n'Pins ----------------------------------------------------------------------
    ConfPort(B, _L_,    // PB.0: ---    : ---
                _L_,    // PB.1: BUZZ   : Signalgeber, Frequenz über Timer 1 PWM, OC1A
                _L_,    // PB.2: ---    : ---
                _L_,    // PB.3: MOSI   : ISP
                _L_,    // PB.4: MISO   : ISP
                _L_,    // PB.5: SCK    : ISP
                _T_,    // PB.6: XTAL1  : Quarz
                _T_);   // PB.7: XTAL2  : Quarz				
    ConfPort(C, _L_,    // PC.0: ---    : ---
                _L_,    // PC.1: Sw_Set : Ausgang, H = Ein
                _T_,    // PC.2: Ct_In  : Kontakt, H = geöffnet
                _L_,    // PC.3: ---    : ---
                _L_,    // PC.4: ---    : ---
                _L_,    // PC.5: ---    : ---
                _T_,    // PC.6: [RESET]: Reset
                _T_);   //				
    ConfPort(D, _T_,    // PD.0: RxD    : µC <- Bus
                _T_,    // PD.1: TxD    : µC -> Bus
                _L_,    // PD.2: ---    : ---
                _L_,    // PD.3: SysLED : System LED, H = Ein
                _L_,    // PD.4: ---    : ---
                _L_,    // PD.5: ---    : ---
                _L_,    // PD.6: [OCR0A]: Externe LED, H = Ein, PWM über Timer 0
                _L_);   // PD.7: ---    : ---
    ConfPort(E, _L_,    // PE.0: ---    : ---
                _L_,    // PE.1: ---    : ---
                _L_,    // PE.2: ---    : ---
                _L_,    // PE.3: ---    : ---
                _L_,    //
                _L_,    //
                _L_,    //
                _L_);   //	
    // Pause, bis sich alles stabilisiert hat --------------------------------------------
    Freeze100us();
	// Timer/Counter0 als Phase Correct PWM an OC0A mit (F_CPU÷64)÷256÷2 = 336Hz
    TCCR0A = BV(COM0A1) | BV(WGM00);
    TCCR0B = BV(CS01) | BV(CS00);
    OCR0A  = 0;
    // Timer/Counter 1 als CTC an OC1A --- (Frequenzgenerator für den Signalgeber) -------
    TCCR1A = 0;
    // Timer 2: Clear Timer on Compare (CTC), Intervall: 1ms (Wiederholrate: ca. 1kHz) ---
    TCCR2A  = BV(WGM21);
    TCCR2B  = BV(CS22) | BV(CS21);
    OCR2A   = F_CPU/256/1000-1;      // = 0.9954 ms
    TIMSK2 |= BV(OCIE2A);
    // USART 19200 Baud, 8 Data, 1 Stop, No Parity, Async., IRQs: RXC/TXC ----------------
    UBRR0  = F_CPU/16L/19200-1;
    UCSR0A = 0;
    UCSR0B = BV(RXCIE0) | BV(TXCIE0) | BV(RXEN0) | BV(TXEN0);
    UCSR0C = BV(UCSZ01) | BV(UCSZ00);
    // Sonstiges -------------------------------------------------------------------------
    EnableInterrupts();
    doNext = DO_PROCEED;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_DoMainLoop(void)
{
    return doNext == DO_PROCEED;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Proceed(void)
{	
	ProceedTonePlaying();
	ProceedSystemTime();
	// Bus-Aktivitätsindikator
	UInt16 ticks = BOARD_MSecTicks();
	{   static UInt08 ms = 0;
		if (ms != (UInt08)ticks)
		{   ms = (UInt08)ticks;
			if (busActivityIndicator > 0)
			{   busActivityIndicator--;
			}
		}
	}	
	// Sonstiges
	wdt_reset();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Shutdown(void)
{
    while( BOARD_BusIsTransmitting() );
    switch( doNext )
    {   case DO_PROCEED:      doNext = DO_SHUTDOWN; return;
        case DO_ENTER_PGMLDR: EnterPgmLdr();        break;
        case DO_STARTUP:      Reset();              break;
        default:              ;
    }
    // Abschaltanweisungen ...
}

//////////////////////////////////////////////////////////////////////////////////////////
