﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Lichtsignale                                                                        //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2017 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include "Conf.h"
#include "Timer.h"
#include "Storing.h"
#include "LightSignal.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef struct              // Licht:
{   UInt08  on;             // · LED-Zustand Ein/Aus = 'true'/'false'
	UInt08  duration;       // · Einschaltdauer: 1 .. 250 [10 ms]
} Light;

typedef struct              // Licht-Sequenz:
{   const Light*  sequence; // · Zeiger auf die Lichtabfolge
	UInt08		  count;    // · Anzahl der Einträge in der Lichtabfolge
} LightSequence;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static Timer            ledOnTimer                = TIMER_MILLISECONDS(0);
static LightEventSignal currentEventSignal        = LIGHTEVENTSIGNAL_NONE;
static UInt08			currentEventBrightness    = 80;
static LightSequence    currentEventSequence      = { 0 };
static UInt08 	        currentEventSequenceIndex = 0;
static LightStateSignal currentStateSignal        = LIGHTSTATESIGNAL_OFF;
static UInt08			currentStateBrightness    = 80;
static LightSequence    currentStateSequence      = { 0 };
static UInt08           currentStateSequenceIndex = 0;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  TransLinear(): Bildet den Wert aus dem Definitionsbereich [srcMin, srcMax] linear   //
//  -------------  und diskret auf den Wertebereich [tgtStart, tgtStop] ab. Liegt       //
//  <value> oberhalb (unterhalb) des Definitionsbereiches, wird tgtStop (tgtStart)      //
//  zurückgegeben. TransLinear() arbeitet ausschliesslich mit Ganzzahlarithmetik.       //
//                                                                                      //
//  Beispiele:                                                                          //
//                                                                                      //
//      TransLinear(value, 3, 14, 6, 3):         TransLinear(value, 2, 5, 3, 9):        //
//                                                                                      //
//                                                       | · · · · · ·                  //
//      | · · · · · · · · · · · · · · · ·               9+ · · · · * *                  //
//     6+ * * * * · · · · · · · · · · · ·                | · · · · · ·                  //
//     5+ · · · · * * * * · · · · · · · ·               7+ · · · * · ·                  //
//     4+ · · · · · · · · * * * * · · · ·                | · · · · · ·                  //
//     3+ · · · · · · · · · · · · * * * *               5+ · · * · · ·                  //
//      | · · · · · · · · · · · · · · · ·                | · · · · · ·                  //
//      | · · | · · · · · · · · · · | · ·               3+ * * · · · ·                  //
//      +-+-+-|-+-+-+-+-+-+-+-+-+-+-|-+-+---->           | · · · · · ·                  //
//     0  1 2 3 4 5 6 7 8 9 1 1 1 1 1 1 1  value         | · | · · | ·                  //
//                          0 1 2 3 4 5 6                +-+-|-+-+-|-+--->              //
//                                                      0  1 2 3 4 5 6   value          //
//                                                                                      //
//  Eingabe: · value:    Der abzubildende Wert                                          //
//           · srcMin:   Definitionsbereich, untere Grenze: -128 .. srcMax              //
//           · srcMax:   Definitionsbereich, obere Grenze:  srcMin .. 127               //
//           · tgtStart: Wertebereich, Anfang: -128 .. 127                              //
//           · tgtStop:  Wertebereich, Ende:   -128 .. 127                              //
//  Ausgabe: Der abgebildete Wert                                                       //
//////////////////////////////////////////////////////////////////////////////////////////
static SInt08
TransLinear(SInt08 value, SInt08 srcMin, SInt08 srcMax, SInt08 tgtStart, SInt08 tgtStop)
{
    if( value <= srcMin ) return tgtStart;
    if( value >= srcMax ) return tgtStop;
    UInt08 srcDelta = (UInt08)(srcMax - srcMin);        // Annahme: srcMin < srcMax
    bool increasing = tgtStart < tgtStop;
    UInt08 dstDelta = increasing ? (UInt08)(tgtStop-tgtStart)
    : (UInt08)(tgtStart-tgtStop);
    UInt16 dividend  = (UInt16)((UInt08)(value-srcMin)) * dstDelta;
    UInt08 divisor   = srcDelta;
    UInt08 quotient  = (UInt08)(dividend / divisor);    // Passt immer in 8 Bit
    UInt08 remainder = (UInt08)(dividend % divisor);
    UInt08 roundGap  = divisor-remainder;
    SInt08 result    = tgtStart;
    if( increasing )
    {   result += quotient;
        if( result < 0 )
        {   if( remainder >  roundGap ) result++;
        }
        else
        {   if( remainder >= roundGap ) result++;
        }
    }
    else
    {   result -= quotient;
        if( result > 0 )
        {   if( remainder >  roundGap ) result--;
        }
        else
        {   if( remainder >= roundGap ) result--;
        }
    }
    return result;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetLight(const LightSequence* sequence, UInt08 index, UInt08 brightness)
{
	Light light;
	memcpy_P(&light, &sequence->sequence[index], sizeof(Light));
	if (light.on) BOARD_SetLed(brightness);
	else	      BOARD_SetLed(0);	
	TIMER_Start(&ledOnTimer, (UInt16)light.duration*10);
}

//////////////////////////////////////////////////////////////////////////////////////////
// Funktionen ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  LIGHTSIGNAL_Play(): Spielt einmalig das in <signal> angegebene Ereignis-Lichtsignal	//
//  ------------------  ab. Eine gegebenenfall laufende Zustands-Lichtsignalausgabe 	//
//	wird für die Abspielzeit des Ereignis-Lichtsignals unterbrochen und anschliessend	//
//  fortgesetzt, d.h. die Ereignis- dominiert die Zustands-Signalisierung. 				//
//                                                                                      //
//  Eingabe: · signal:     Das auszugebende Ereignis-Lichtsignal gem. Aufzählungstyp 	//
//                         <LightEventSignal>.											//
//           · brightness: Helligkeit des Lichtsignals: 10 .. 100%.						//
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
LIGHTSIGNAL_Play(LightEventSignal signal, UInt08 brightness)
{
	// Ereignis-Lichtsignale (einmalig ablaufende Lichtsignalabfolgen)
	static const Light sequence01[] PROGMEM = { { true, 25 }, { false, 25 } };
	static const Light sequence02[] PROGMEM = { { true, 25 }, { false, 25 }, { true, 25 }, { false, 25 } };		
	static const Light sequence03[] PROGMEM = { { true, 25 }, { false, 25 }, { true, 25 }, { false, 25 }, { true, 25 }, { false, 25 } };
	static const Light sequence04[] PROGMEM = { { true, 25 }, { false, 25 }, { true, 25 }, { false, 25 }, { true, 25 }, { false, 25 }, { true, 25 }, { false, 25 }, { true, 25 }, { false, 25 }, { true, 25 }, { false, 25 } };
	static const Light sequence05[] PROGMEM = { { true,  5 }, { false, 50 } };
	static const Light sequence06[] PROGMEM = { { true,  5 }, { false, 50 }, { true,  5 }, { false, 50 } };
	static const Light sequence07[] PROGMEM = { { true,  5 }, { false, 50 }, { true,  5 }, { false, 50 }, { true,  5 }, { false, 50 } };
	static const Light sequence08[] PROGMEM = { { true,  5 }, { false, 50 }, { true,  5 }, { false, 50 }, { true,  5 }, { false, 50 }, { true,  5 }, { false, 50 }, { true,  5 }, { false, 50 }, { true,  5 }, { false, 50 } };
	static const Light sequence11[] PROGMEM = { { true, 10 }, { false, 10 }, { true, 10 }, { false, 10 }, { true, 10 }, { false, 10 }, { true, 10 }, { false, 10 }, { true, 10 }, { false, 10 }, { true, 10 }, { false, 10 } };

	static const LightSequence const table[] PROGMEM =
	{	{ sequence01, CARDINALITY(sequence01) }, // 'LIGHTSIGNAL_BLINK_1X
		{ sequence02, CARDINALITY(sequence02) }, // 'LIGHTSIGNAL_BLINK_2X
    	{ sequence03, CARDINALITY(sequence03) }, // 'LIGHTSIGNAL_BLINK_3X
    	{ sequence04, CARDINALITY(sequence04) }, // 'LIGHTSIGNAL_BLINK_6X
    	{ sequence05, CARDINALITY(sequence05) }, // 'LIGHTSIGNAL_FLASH_1X
		{ sequence06, CARDINALITY(sequence06) }, // 'LIGHTSIGNAL_BLINK_2X
    	{ sequence07, CARDINALITY(sequence07) }, // 'LIGHTSIGNAL_FLASH_3X
    	{ sequence08, CARDINALITY(sequence08) }, // 'LIGHTSIGNAL_FLASH_6X 	
		{ NULL      , 0                       }, // 'LIGHTSIGNAL_FADE_IN' und 
		{ NULL      , 0                       }, // 'LIGHTSIGNAL_FADE_OUT' werden gesondert behandelt
		{ sequence11, CARDINALITY(sequence11) }  // 'LIGHTSIGNAL_FLICKER'
	};

	UInt08 index = (UInt08)signal-1;
	if (signal != LIGHTEVENTSIGNAL_NONE && index < CARDINALITY(table))
	{	memcpy_P(&currentEventSequence, &table[index], sizeof(LightSequence));
		currentEventSequenceIndex = 0;
	}
	else
	{	signal = LIGHTEVENTSIGNAL_NONE;
	}	
	currentEventSignal     = signal;
	currentEventBrightness = brightness;
	TIMER_Start(&ledOnTimer, 0);	
}

//////////////////////////////////////////////////////////////////////////////////////////
//  LIGHTSIGNAL_Start(): Startet die in Dauerschleife laufende Zustands-Lichtsignal-	//
//  -------------------  ausgabe in <signal>. Findet aktuell eine Ereignis-Licht-		//
//	signalausgabe statt (s. LIGHTSIGNAL_Play()) wird die Zustands-Lichtausgabe erst		//
//	nach Ablauf der Ereignis-Lichtsignalausgabe gestartet.								//
//                                                                                      //
//  Eingabe: · signal:     Das auszugebende Zustands-Lichtsignal gem. Aufzählungstyp	//
//                         <LightEventSignal>.											//
//           · brightness: Helligkeit des Lichtsignals: 10 .. 100%.						//
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
LIGHTSIGNAL_Start(LightStateSignal signal, UInt08 brightness)
{
	// Zustands-Lichtsignale (in Dauerschleife wiederholte Lichtsignalabfolge)
	static const Light sequence01[] PROGMEM = { { true,  5 } };
	static const Light sequence02[] PROGMEM = { { true, 25 }, { false,  25 } };
	static const Light sequence03[] PROGMEM = { { true,  5 }, { false, 100 } };		
	
	static const LightSequence const table[] PROGMEM =
	{	{ sequence01, CARDINALITY(sequence01) },	// 'LIGHTSTATESIGNAL_ON'
		{ sequence02, CARDINALITY(sequence02) },	// 'LIGHTSTATESIGNAL_BLINK'
		{ sequence03, CARDINALITY(sequence03) },	// 'LIGHTSTATESIGNAL_FLASH'
		{ NULL      , 0                       },	// 'LIGHTSTATESIGNAL_BREATHE'   (wird gesondert behandelt)
		{ NULL      , 0                       }		// 'LIGHTSTATESIGNAL_HEARTBEAT' (wird gesondert behandelt)
	};

	UInt08 index = (UInt08)signal-1;
	if (signal != LIGHTSTATESIGNAL_OFF && index < CARDINALITY(table))
	{	memcpy_P(&currentStateSequence, &table[index], sizeof(LightSequence));
		currentStateSequenceIndex = 0;
	}
	else
	{	signal = LIGHTSTATESIGNAL_OFF;
	}
	currentStateSignal     = signal;
	currentStateBrightness = brightness;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  LIGHTSIGNAL_Current(): Liefert die aktuell laufende Zustands-Lichtsignalausgabe.    //
//  --------------------																//
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Die aktuelle Zustands-Lichtsignalausgabe als <LightStateSignal>.           //
//////////////////////////////////////////////////////////////////////////////////////////
LightStateSignal
LIGHTSIGNAL_Current(void)
{
	return currentStateSignal;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  LIGHTSIGNAL_Brightness(): Liefert den Helligkeitswert, der beim	letzten Aufruf von	//
//  ------------------------  LIGHTSIGNAL_Start() mitgegeben wurde.						//
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Die aktuelle Helligkeit der Zustands-Lichtsignalausgabe: 10 .. 100 %.		//
//////////////////////////////////////////////////////////////////////////////////////////
UInt08
LIGHTSIGNAL_Brightness(void)
{
	return currentStateBrightness;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  LIGHTSIGNAL_Stop(): Beendet die laufende Zustands-Lichtsignalausgabe.			    //
//  ------------------																	//
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
LIGHTSIGNAL_Stop(void)
{
	LIGHTSIGNAL_Start(LIGHTSTATESIGNAL_OFF, 0);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
LIGHTSIGNAL_Startup(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
void
LIGHTSIGNAL_Proceed(void)
{		
	static bool   rising     = true;
	static UInt08 counter    = 0;	// %    
	    
	// Anzeige der Busaktivität (hat Vorrang gegenüber der Lichtsignalisierung)
	LedBusActivity ledBusActivity;
	STORING_Load(STO_LedBusActivity, &ledBusActivity);
	if (ledBusActivity.enabled && BOARD_BusActivity())
	{	BOARD_SetLed(ledBusActivity.brightness);
		return;
	}
	// Lichtsignalisierung
	if (!TIMER_HasExpired(&ledOnTimer)) return;
	if (currentEventSignal != LIGHTEVENTSIGNAL_NONE)
	{	// Ereignis-Lichtsignalausgabe        
		if (currentEventSignal == LIGHTSIGNAL_FADE_IN)
		{	UInt08 brightness = (UInt08)TransLinear(counter, 0, 100, 0, currentEventBrightness);
            BOARD_SetLed(brightness);
			TIMER_Start(&ledOnTimer, 10);            
			if (counter == 100) currentEventSignal = LIGHTEVENTSIGNAL_NONE;
			else                counter++;
		}
		else if (currentEventSignal == LIGHTSIGNAL_FADE_OUT)
		{	UInt08 brightness = (UInt08)TransLinear(counter, 0, 100, 0, currentEventBrightness);            
			if (counter == 0) counter = currentEventBrightness;
			BOARD_SetLed(brightness);
			TIMER_Start(&ledOnTimer, 10);		
			counter--;	
			if (counter == 0) currentEventSignal = LIGHTEVENTSIGNAL_NONE;			
		}
		else
		{	if (currentEventSequenceIndex == currentEventSequence.count)
			{	currentEventSignal = LIGHTEVENTSIGNAL_NONE;
				BOARD_SetLed(0);
				rising  = true;
				counter = 0;	// %				
			}
			else
			{	SetLight(&currentEventSequence, currentEventSequenceIndex, currentEventBrightness);
				currentEventSequenceIndex++;
			}
		}
	}
	else if (currentStateSignal != LIGHTSTATESIGNAL_OFF)
	{	// Zustands-Lichtsignalausgabe		
		if (currentStateSignal == LIGHTSTATESIGNAL_HEARTBEAT)
		{	UInt16 ticks = BOARD_MSecTicks();
			bool ledOn = (bool)((ticks & 0x380)==0x80 || (ticks & 0x380)==0x180);
			if (ledOn) BOARD_SetLed(currentStateBrightness);
			else       BOARD_SetLed(0);
			rising  = true;
			counter = 0;	// %
		}
		else if (currentStateSignal == LIGHTSTATESIGNAL_PULSE)
		{	if (rising)
			{	if (counter == 100) rising = false;
				else                counter++;
			}
			else
			{	
				if (counter == 0) rising = true;
				else              counter--;
			}						
			UInt08 brightness = (UInt08)TransLinear(counter, 0, 100, 0, currentStateBrightness);
			BOARD_SetLed(brightness);
			TIMER_Start(&ledOnTimer, 10);			 			 
		}
		else if (currentStateSequenceIndex == currentStateSequence.count)
		{	currentStateSequenceIndex = 0;			
		}
		else
		{	SetLight(&currentStateSequence, currentStateSequenceIndex, currentStateBrightness);
			currentStateSequenceIndex++;
		}						
	}
	else
	{	BOARD_SetLed(0);
		rising  = true;
		counter = 0;	// %
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
void
LIGHTSIGNAL_Shutdown(void)
{
    LIGHTSIGNAL_Stop();
}

//////////////////////////////////////////////////////////////////////////////////////////