﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Tonsignale                                                                          //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2017 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include "Conf.h"
#include "Board.h"
#include "Timer.h"
#include "ToneSignal.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef struct              // Ton:
{   Note    pitch;          // · Tonhöhe
    UInt08  sustain;        // · Haltedauer: 1 .. 250 [10 ms]
} Tone;

typedef struct              // Ton-Sequenz:
{   const Tone*   sequence; // · Zeiger auf die Toneabfolge
	UInt08		  count;    // · Anzahl der Töne in Toneabfolge 
} ToneSequence;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static ToneEventSignal currentEventSignal        = TONEEVENTSIGNAL_NONE;
static ToneSequence    currentEventSequence      = { 0 };
static UInt08 	       currentEventSequenceIndex = 0;
static ToneStateSignal currentStateSignal        = TONESTATESIGNAL_OFF;
static ToneSequence    currentStateSequence      = { 0 };
static UInt08          currentStateSequenceIndex = 0;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
PlayTone(const ToneSequence* sequence, UInt08 index)
{
	Tone tone;
	memcpy_P(&tone, &sequence->sequence[index], sizeof(Tone));
	BOARD_PlayTone(tone.pitch, tone.sustain);
}

//////////////////////////////////////////////////////////////////////////////////////////
// Funktionen ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  TONESIGNAL_Play(): Spielt einmalig das in <signal> angegebene Ereignis-Tonsignal	//
//  -----------------  ab. Eine gegebenenfall laufende Zustands-Tonsignalausgabe wird	//
//	für die Abspielzeit des Ereignis-Tonsignals unterbrochen und anschliessend fort-    //
//	gesetzt, d.h. die Ereignis- dominiert die Zustands-Signalisierung.					//
//                                                                                      //
//  Eingabe: · signal: Das auszugebende Ereignis-Tonsignal gem. Aufzählungstyp <Tone-	//
//                     EventSignal>.													//
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
TONESIGNAL_Play(ToneEventSignal signal)
{		
	// Ereignis-Tonsignale (einmalig ablaufende Tonabfolgen) 
	static const Tone sequence01[] PROGMEM = { { NOTE_C6,  7 }, { NOTE_OFF,  6 }, { NOTE_C8, 30 } };
	static const Tone sequence02[] PROGMEM = { { NOTE_C8,  7 }, { NOTE_OFF,  6 }, { NOTE_C6, 30 } };
	static const Tone sequence03[] PROGMEM = { { NOTE_C6,  8 }, { NOTE_G7,   8 }, { NOTE_C8, 40 } };			
	static const Tone sequence04[] PROGMEM = { { NOTE_C8,  8 }, { NOTE_G7,   8 }, { NOTE_C5, 40 } };	
	static const Tone sequence05[] PROGMEM = { { NOTE_A5, 10 }, { NOTE_A7,  20 } };
	static const Tone sequence06[] PROGMEM = { { NOTE_C6, 10 }, { NOTE_OFF, 50 }, { NOTE_C8, 10 } };
	static const Tone sequence07[] PROGMEM = { { NOTE_C8,  5 }, { NOTE_OFF,  5 }, { NOTE_C8, 20 } };
	static const Tone sequence08[] PROGMEM = { { NOTE_C8,  5 }, { NOTE_OFF,  5 }, { NOTE_C8,  5 }, { NOTE_OFF,  5 }, { NOTE_C8, 20 } };
	static const Tone sequence09[] PROGMEM = { { NOTE_C8,  5 }, { NOTE_OFF,  5 }, { NOTE_F8,  5 }, { NOTE_OFF,  5 }, { NOTE_C8, 20 } };
	static const Tone sequence10[] PROGMEM = { { NOTE_C6,  5 } };
	static const Tone sequence11[] PROGMEM = { { NOTE_C6, 40 } };
	static const Tone sequence12[] PROGMEM = { { NOTE_C8,  5 } };
	static const Tone sequence13[] PROGMEM = { { NOTE_C8, 40 } };
	static const Tone sequence14[] PROGMEM = { { NOTE_C9,  5 } };
	static const Tone sequence15[] PROGMEM = { { NOTE_C9, 40 } };		
	
	static const ToneSequence const table[] PROGMEM =
	{	{ sequence01, CARDINALITY(sequence01) }, // · Tief Hoch
	    { sequence02, CARDINALITY(sequence02) }, // · Hoch Tief
	    { sequence03, CARDINALITY(sequence03) }, // · Tief Mitte Hoch
	    { sequence04, CARDINALITY(sequence04) }, // · Hoch Mitte Tief
	    { sequence05, CARDINALITY(sequence05) }, // · Taste gedrückt
	    { sequence06, CARDINALITY(sequence06) }, // · C6 zu C8
	    { sequence07, CARDINALITY(sequence07) }, // · Doppelton
	    { sequence08, CARDINALITY(sequence08) }, // · Trippelton
	    { sequence09, CARDINALITY(sequence09) }, // · Zwitscher
	    { sequence10, CARDINALITY(sequence10) }, // · Kurzer Piepton C6
	    { sequence11, CARDINALITY(sequence11) }, // · Langer Piepton C6
		{ sequence12, CARDINALITY(sequence12) }, // · Kurzer Piepton C8
		{ sequence13, CARDINALITY(sequence13) }, // · Langer Piepton C8			
		{ sequence14, CARDINALITY(sequence14) }, // · Kurzer Piepton C9
		{ sequence15, CARDINALITY(sequence15) }	 // · Langer Piepton C9			
	};
		
	UInt08 index = (UInt08)signal-1;
	if (signal != TONEEVENTSIGNAL_NONE && index < CARDINALITY(table))
	{	memcpy_P(&currentEventSequence, &table[index], sizeof(ToneSequence));		
		currentEventSequenceIndex = 0;		
	}
	else
	{	signal = TONEEVENTSIGNAL_NONE;
	}		
	currentEventSignal = signal;
	BOARD_PlayTone(NOTE_OFF, 0);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TONESIGNAL_Start(): Startet die in Dauerschleife laufende Zustands-Tonsignalausgabe	//
//  ------------------  in <signal>. Findet aktuell eine Ereignis-Tonsignalausgabe  	//
//	statt (s. TONESIGNAL_Play()) wird die Zustands-Tonausgabe erst nach Ablauf der		//
//	Ereignis-Tonsignalausgabe gestartet.												//
//                                                                                      //
//  Eingabe: · signal: Das auszugebende Zustands-Tonsignal gem. Aufzählungstyp <Tone-	//
//					   EventSignal>.													//
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
TONESIGNAL_Start(ToneStateSignal signal)
{	
	// Zustands-Tonsignale (in Dauerschleife wiederholte Tonabfolgen)
	static const Tone sequence01[] PROGMEM = { { NOTE_C8,   3 }, { NOTE_OFF, 150 } };
	static const Tone sequence02[] PROGMEM = { { NOTE_A8,   2 }, { NOTE_OFF, 200 } };
	static const Tone sequence03[] PROGMEM = { { NOTE_D7,   5 }, { NOTE_OFF, 100 }, { NOTE_A5, 5 }, { NOTE_OFF, 100 } };
	static const Tone sequence04[] PROGMEM = { { NOTE_D7,  30 }, { NOTE_OFF,  30 } };
	static const Tone sequence05[] PROGMEM = { { NOTE_D7, 100 }, { NOTE_OFF, 100 } };
				
	static const ToneSequence const table[] PROGMEM =
	{	{ sequence01, CARDINALITY(sequence01) }, // · Tropfen
	    { sequence02, CARDINALITY(sequence02) }, // · Ticken
	    { sequence03, CARDINALITY(sequence03) }, // · Martinticks
	    { sequence04, CARDINALITY(sequence04) }, // · Kurzer Piepton D7
	    { sequence05, CARDINALITY(sequence05) }	 // · Langer Piepton D7
	};
	
	UInt08 index = (UInt08)signal-1;
	if (signal != TONESTATESIGNAL_OFF && index < CARDINALITY(table))
	{	memcpy_P(&currentStateSequence, &table[index], sizeof(ToneSequence));
		currentStateSequenceIndex = 0;
	}
	else
	{	signal = TONESTATESIGNAL_OFF;		
	}	
	currentStateSignal = signal;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TONESIGNAL_Current(): Liefert die aktuell laufende Zustands-Tonsignalausgabe.	    //
//  --------------------																//
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Die aktuelle Zustands-Tonsignalausgabe als <ToneStateSignal>.              //
//////////////////////////////////////////////////////////////////////////////////////////
ToneStateSignal
TONESIGNAL_Current(void)
{
	return currentStateSignal;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TONESIGNAL_Stop(): Beendet die laufende Zustands-Tonsignalausgabe.					//
//  -----------------																	//
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
TONESIGNAL_Stop(void)
{
    TONESIGNAL_Start(TONESTATESIGNAL_OFF);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
TONESIGNAL_Startup(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
void
TONESIGNAL_Proceed(void)
{
	if (!BOARD_IsSilence()) return;	
	if (currentEventSignal != TONEEVENTSIGNAL_NONE)	
	{	// Ereignis-Tonesignalausgabe
		if (currentEventSequenceIndex == currentEventSequence.count)
		{	currentEventSignal = TONEEVENTSIGNAL_NONE;
			BOARD_PlayTone(NOTE_OFF, 25);
		}
		else		
		{	PlayTone(&currentEventSequence, currentEventSequenceIndex);
			currentEventSequenceIndex++;
		}		
	}	
	else if (currentStateSignal != TONESTATESIGNAL_OFF)
	{	// Zustands-Tonsignalausgabe
		if (currentStateSequenceIndex == currentStateSequence.count)
		{	currentStateSequenceIndex = 0;				
			BOARD_PlayTone(NOTE_OFF, 0);
		}
		else
		{	PlayTone(&currentStateSequence, currentStateSequenceIndex);
			currentStateSequenceIndex++;
		}
	}	
}

//////////////////////////////////////////////////////////////////////////////////////////
void
TONESIGNAL_Shutdown(void)
{	
    TONESIGNAL_Stop();
}

//////////////////////////////////////////////////////////////////////////////////////////